#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
import re
import warnings

from IPython.display import display
from pandas.tseries.offsets import CustomBusinessDay, Day, BusinessDay
from scipy.stats import mode


class NonMatchingTimezoneError(Exception):
    pass


class MaxLossExceededError(Exception):
    pass


def rethrow(exception, additional_message):
    """
    Re-raise the last exception that was active in the current scope
    without losing the stacktrace but adding an additional message.
    This is hacky because it has to be compatible with both python 2/3
    """
    e = exception
    m = additional_message
    if not e.args:
        e.args = (m,)
    else:
        e.args = (e.args[0] + m,) + e.args[1:]
    raise e


def non_unique_bin_edges_error(func):
    """
    Give user a more informative error in case it is not possible
    to properly calculate quantiles on the input dataframe (factor)
    """
    message = """

    An error occurred while computing bins/quantiles on the input provided.
    This usually happens when the input contains too many identical
    values and they span more than one quantile. The quantiles are choosen
    to have the same number of records each, but the same value cannot span
    multiple quantiles. Possible workarounds are:
    1 - Decrease the number of quantiles
    2 - Specify a custom quantiles range, e.g. [0, .50, .75, 1.] to get unequal
        number of records per quantile
    3 - Use 'bins' option instead of 'quantiles', 'bins' chooses the
        buckets to be evenly spaced according to the values themselves, while
        'quantiles' forces the buckets to have the same number of records.
    4 - for factors with discrete values use the 'bins' option with custom
        ranges and create a range for each discrete value
    Please see utils.get_clean_factor_and_forward_returns documentation for
    full documentation of 'bins' and 'quantiles' options.

"""

    def dec(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValueError as e:
            if 'Bin edges must be unique' in str(e):
                rethrow(e, message)
            raise
    return dec


@non_unique_bin_edges_error
def quantize_factor(factor_data,
                    quantiles=5,
                    bins=None,
                    by_group=False,
                    no_raise=False,
                    zero_aware=False):
    """
    Computes period wise factor quantiles.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.

        - See full explanation in utils.get_clean_factor_and_forward_returns

    quantiles : int or sequence[float]
        Number of equal-sized quantile buckets to use in factor bucketing.
        Alternately sequence of quantiles, allowing non-equal-sized buckets
        e.g. [0, .10, .5, .90, 1.] or [.05, .5, .95]
        Only one of 'quantiles' or 'bins' can be not-None
    bins : int or sequence[float]
        Number of equal-width (valuewise) bins to use in factor bucketing.
        Alternately sequence of bin edges allowing for non-uniform bin width
        e.g. [-4, -2, -0.5, 0, 10]
        Only one of 'quantiles' or 'bins' can be not-None
    by_group : bool, optional
        If True, compute quantile buckets separately for each group.
    no_raise: bool, optional
        If True, no exceptions are thrown and the values for which the
        exception would have been thrown are set to np.NaN
    zero_aware : bool, optional
        If True, compute quantile buckets separately for positive and negative
        signal values. This is useful if your signal is centered and zero is
        the separation between long and short signals, respectively.

    Returns
    -------
    factor_quantile : pd.Series
        Factor quantiles indexed by date and asset.
    """
    if not ((quantiles is not None and bins is None) or
            (quantiles is None and bins is not None)):
        raise ValueError('Either quantiles or bins should be provided')

    if zero_aware and not (isinstance(quantiles, int)
                           or isinstance(bins, int)):
        msg = ("zero_aware should only be True when quantiles or bins is an"
               " integer")
        raise ValueError(msg)

    def quantile_calc(x, _quantiles, _bins, _zero_aware, _no_raise):
        try:
            if _quantiles is not None and _bins is None and not _zero_aware:
                return pd.qcut(x, _quantiles, labels=False) + 1
            elif _quantiles is not None and _bins is None and _zero_aware:
                pos_quantiles = pd.qcut(x[x >= 0], _quantiles // 2,
                                        labels=False) + _quantiles // 2 + 1
                neg_quantiles = pd.qcut(x[x < 0], _quantiles // 2,
                                        labels=False) + 1
                return pd.concat([pos_quantiles, neg_quantiles]).sort_index()
            elif _bins is not None and _quantiles is None and not _zero_aware:
                return pd.cut(x, _bins, labels=False) + 1
            elif _bins is not None and _quantiles is None and _zero_aware:
                pos_bins = pd.cut(x[x >= 0], _bins // 2,
                                  labels=False) + _bins // 2 + 1
                neg_bins = pd.cut(x[x < 0], _bins // 2,
                                  labels=False) + 1
                return pd.concat([pos_bins, neg_bins]).sort_index()
        except Exception as e:
            if _no_raise:
                return pd.Series(index=x.index)
            raise e

    grouper = [factor_data.index.get_level_values('date')]
    if by_group:
        grouper.append('group')

    factor_quantile = factor_data.groupby(grouper,group_keys=False)['factor'] \
        .apply(quantile_calc, quantiles, bins, zero_aware, no_raise)
    factor_quantile.name = 'factor_quantile'

    return factor_quantile.dropna()


def infer_trading_calendar(factor_idx, prices_idx):
    """
    Infer the trading calendar from factor and price information.

    Parameters
    ----------
    factor_idx : pd.DatetimeIndex
        The factor datetimes for which we are computing the forward returns
    prices_idx : pd.DatetimeIndex
        The prices datetimes associated withthe factor data

    Returns
    -------
    calendar : pd.DateOffset
    """
    full_idx = factor_idx.union(prices_idx)

    traded_weekdays = []
    holidays = []

    days_of_the_week = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    for day, day_str in enumerate(days_of_the_week):

        weekday_mask = (full_idx.dayofweek == day)

        # drop days of the week that are not traded at all
        if not weekday_mask.any():
            continue
        traded_weekdays.append(day_str)

        # look for holidays
        used_weekdays = full_idx[weekday_mask].normalize()
        all_weekdays = pd.date_range(full_idx.min(), full_idx.max(),
                                     freq=CustomBusinessDay(weekmask=day_str)
                                     ).normalize()
        _holidays = all_weekdays.difference(used_weekdays)
        _holidays = [timestamp.date() for timestamp in _holidays]
        holidays.extend(_holidays)

    traded_weekdays = ' '.join(traded_weekdays)
    return CustomBusinessDay(weekmask=traded_weekdays, holidays=holidays)


def compute_forward_returns(factor,
                            prices,
                            periods=(1, 5, 10),
                            filter_zscore=None,
                            cumulative_returns=True,
                            periods_by_factor=False):
    """
    Finds the N period forward returns (as percent change) for each asset
    provided.

    Parameters
    ----------
    factor : pd.Series - MultiIndex
        A MultiIndex Series indexed by timestamp (level 0) and asset
        (level 1), containing the values for a single alpha factor.

        - See full explanation in utils.get_clean_factor_and_forward_returns

    prices : pd.DataFrame
        Pricing data to use in forward price calculation.
        Assets as columns, dates as index. Pricing data must
        span the factor analysis time period plus an additional buffer window
        that is greater than the maximum number of expected periods
        in the forward returns calculations.
    periods : sequence[int]
        periods to compute forward returns on.
    filter_zscore : int or float, optional
        Sets forward returns greater than X standard deviations
        from the the mean to nan. Set it to 'None' to avoid filtering.
        Caution: this outlier filtering incorporates lookahead bias.
    cumulative_returns : bool, optional
        If True, forward returns columns will contain cumulative returns.
        Setting this to False is useful if you want to analyze how predictive
        a factor is for a single forward day.

    Returns
    -------
    forward_returns : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by timestamp (level 0) and asset
        (level 1), containing the forward returns for assets.
        Forward returns column names follow the format accepted by
        pd.Timedelta (e.g. '1D', '30m', '3h15m', '1D1h', etc).
        'date' index freq property (forward_returns.index.levels[0].freq)
        will be set to a trading calendar (pandas DateOffset) inferred
        from the input data (see infer_trading_calendar for more details).
    """

    factor_dateindex = factor.index.levels[0]
    if factor_dateindex.tz != prices.index.tz:
        raise NonMatchingTimezoneError("The timezone of 'factor' is not the "
                                       "same as the timezone of 'prices'. See "
                                       "the pandas methods tz_localize and "
                                       "tz_convert.")

    freq = infer_trading_calendar(factor_dateindex, prices.index)

    factor_dateindex = factor_dateindex.intersection(prices.index)

    if len(factor_dateindex) == 0:
        raise ValueError("Factor and prices indices don't match: make sure "
                         "they have the same convention in terms of datetimes "
                         "and symbol-names")

    # chop prices down to only the assets we care about (= unique assets in
    # `factor`).  we could modify `prices` in place, but that might confuse
    # the caller.
    prices = prices.filter(items=factor.index.levels[1])

    raw_values_dict = {}
    column_list = []


    if not periods_by_factor:
        for period in sorted(periods):
            if cumulative_returns:
                returns = prices.pct_change(period)
            else:
                returns = prices.pct_change()
            

            forward_returns = \
                returns.shift(-period).reindex(factor_dateindex)

            if filter_zscore is not None:
                mask = abs(
                    forward_returns - forward_returns.mean()
                ) > (filter_zscore * forward_returns.std())
                forward_returns[mask] = np.nan

            #
            # Find the period length, which will be the column name. We'll test
            # several entries in order to find out the most likely period length
            # (in case the user passed inconsinstent data)
            #
            days_diffs = []
            for i in range(30):
                if i >= len(forward_returns.index):
                    break
                p_idx = prices.index.get_loc(forward_returns.index[i])
                if p_idx is None or p_idx < 0 or (
                        p_idx + period) >= len(prices.index):
                    continue
                start = prices.index[p_idx]
                end = prices.index[p_idx + period]
                period_len = diff_custom_calendar_timedeltas(start, end, freq)
                days_diffs.append(period_len.components.days)

            delta_days = period_len.components.days - mode(days_diffs,keepdims=True).mode[0]
            period_len -= pd.Timedelta(days=delta_days)
            label = timedelta_to_string(period_len)




            column_list.append(label)

            raw_values_dict[label] = np.concatenate(forward_returns.values)
    else:
        prices = prices.loc[factor_dateindex]
        returns = prices.pct_change().fillna(0)
        forward_returns = returns.shift(-1).reindex(factor_dateindex)
        
        if filter_zscore is not None:
            mask = abs(
                forward_returns - forward_returns.mean()
            ) > (filter_zscore * forward_returns.std())
            forward_returns[mask] = np.nan
        
        label = '1T'
        column_list.append(label)
        raw_values_dict[label] = np.concatenate(forward_returns.values)



    df = pd.DataFrame.from_dict(raw_values_dict)
    df.set_index(
        pd.MultiIndex.from_product(
            [factor_dateindex, prices.columns],
            names=['date', 'asset']
        ),
        inplace=True
    )
    df = df.reindex(factor.index)

    # now set the columns correctly
    df = df[column_list]
    # if periods_by_factor:
    #     df.index.levels[0].freq = None 
    # else:
        # df.index.levels[0].freq = freq
    df.index.set_names(['date', 'asset'], inplace=True)

    return df


def backshift_returns_series(series, N):
    """Shift a multi-indexed series backwards by N observations in
    the first level.

    This can be used to convert backward-looking returns into a
    forward-returns series.
    """
    ix = series.index
    dates, sids = ix.levels
    date_labels, sid_labels = map(np.array, ix.labels)

    # Output date labels will contain the all but the last N dates.
    new_dates = dates[:-N]

    # Output data will remove the first M rows, where M is the index of the
    # last record with one of the first N dates.
    cutoff = date_labels.searchsorted(N)
    new_date_labels = date_labels[cutoff:] - N
    new_sid_labels = sid_labels[cutoff:]
    new_values = series.values[cutoff:]

    assert new_date_labels[0] == 0

    new_index = pd.MultiIndex(
        levels=[new_dates, sids],
        labels=[new_date_labels, new_sid_labels],
        sortorder=1,
        names=ix.names,
    )

    return pd.Series(data=new_values, index=new_index)


def demean_forward_returns(factor_data, grouper=None):
    """
    Convert forward returns to returns relative to mean
    period wise all-universe or group returns.
    group-wise normalization incorporates the assumption of a
    group neutral portfolio constraint and thus allows allows the
    factor to be evaluated across groups.

    For example, if AAPL 5 period return is 0.1% and mean 5 period
    return for the Technology stocks in our universe was 0.5% in the
    same period, the group adjusted 5 period return for AAPL in this
    period is -0.4%.

    Parameters
    ----------
    factor_data : pd.DataFrame - MultiIndex
        Forward returns indexed by date and asset.
        Separate column for each forward return window.
    grouper : list
        If True, demean according to group.

    Returns
    -------
    adjusted_forward_returns : pd.DataFrame - MultiIndex
        DataFrame of the same format as the input, but with each
        security's returns normalized by group.
    """

    factor_data = factor_data.copy()

    if not grouper:
        grouper = factor_data.index.get_level_values('date')

    cols = get_forward_returns_columns(factor_data.columns)
    factor_data[cols] = factor_data.groupby(grouper)[cols] \
        .transform(lambda x: x - x.mean())

    return factor_data


def print_table(table, name=None, fmt=None):
    """
    Pretty print a pandas DataFrame.

    Uses HTML output if running inside Jupyter Notebook, otherwise
    formatted text output.

    Parameters
    ----------
    table : pd.Series or pd.DataFrame
        Table to pretty-print.
    name : str, optional
        Table name to display in upper left corner.
    fmt : str, optional
        Formatter to use for displaying table elements.
        E.g. '{0:.2f}%' for displaying 100 as '100.00%'.
        Restores original setting after displaying.
    """
    if isinstance(table, pd.Series):
        table = pd.DataFrame(table)

    if isinstance(table, pd.DataFrame):
        table.columns.name = name

    prev_option = pd.get_option('display.float_format')
    if fmt is not None:
        pd.set_option('display.float_format', lambda x: fmt.format(x))

    display(table)

    if fmt is not None:
        pd.set_option('display.float_format', prev_option)


def get_clean_factor(factor,
                     forward_returns,
                     groupby=None,
                     binning_by_group=False,
                     quantiles=5,
                     bins=None,
                     groupby_labels=None,
                     max_loss=0.35,
                     zero_aware=False):
    """
    Formats the factor data, forward return data, and group mappings into a
    DataFrame that contains aligned MultiIndex indices of timestamp and asset.
    The returned data will be formatted to be suitable for Alphalens functions.

    It is safe to skip a call to this function and still make use of Alphalens
    functionalities as long as the factor data conforms to the format returned
    from get_clean_factor_and_forward_returns and documented here

    Parameters
    ----------
    factor : pd.Series - MultiIndex
        A MultiIndex Series indexed by timestamp (level 0) and asset
        (level 1), containing the values for a single alpha factor.
        ::
            -----------------------------------
                date    |    asset   |
            -----------------------------------
                        |   AAPL     |   0.5
                        -----------------------
                        |   BA       |  -1.1
                        -----------------------
            2014-01-01  |   CMG      |   1.7
                        -----------------------
                        |   DAL      |  -0.1
                        -----------------------
                        |   LULU     |   2.7
                        -----------------------

    forward_returns : pd.DataFrame - MultiIndex
        A MultiIndex DataFrame indexed by timestamp (level 0) and asset
        (level 1), containing the forward returns for assets.
        Forward returns column names must follow the format accepted by
        pd.Timedelta (e.g. '1D', '30m', '3h15m', '1D1h', etc).
        'date' index freq property must be set to a trading calendar
        (pandas DateOffset), see infer_trading_calendar for more details.
        This information is currently used only in cumulative returns
        computation
        ::
            ---------------------------------------
                       |       | 1D  | 5D  | 10D
            ---------------------------------------
                date   | asset |     |     |
            ---------------------------------------
                       | AAPL  | 0.09|-0.01|-0.079
                       ----------------------------
                       | BA    | 0.02| 0.06| 0.020
                       ----------------------------
            2014-01-01 | CMG   | 0.03| 0.09| 0.036
                       ----------------------------
                       | DAL   |-0.02|-0.06|-0.029
                       ----------------------------
                       | LULU  |-0.03| 0.05|-0.009
                       ----------------------------

    groupby : pd.Series - MultiIndex or dict
        Either A MultiIndex Series indexed by date and asset,
        containing the period wise group codes for each asset, or
        a dict of asset to group mappings. If a dict is passed,
        it is assumed that group mappings are unchanged for the
        entire time period of the passed factor data.
    binning_by_group : bool
        If True, compute quantile buckets separately for each group.
        This is useful when the factor values range vary considerably
        across gorups so that it is wise to make the binning group relative.
        You should probably enable this if the factor is intended
        to be analyzed for a group neutral portfolio
    quantiles : int or sequence[float]
        Number of equal-sized quantile buckets to use in factor bucketing.
        Alternately sequence of quantiles, allowing non-equal-sized buckets
        e.g. [0, .10, .5, .90, 1.] or [.05, .5, .95]
        Only one of 'quantiles' or 'bins' can be not-None
    bins : int or sequence[float]
        Number of equal-width (valuewise) bins to use in factor bucketing.
        Alternately sequence of bin edges allowing for non-uniform bin width
        e.g. [-4, -2, -0.5, 0, 10]
        Chooses the buckets to be evenly spaced according to the values
        themselves. Useful when the factor contains discrete values.
        Only one of 'quantiles' or 'bins' can be not-None
    groupby_labels : dict
        A dictionary keyed by group code with values corresponding
        to the display name for each group.
    max_loss : float, optional
        Maximum percentage (0.00 to 1.00) of factor data dropping allowed,
        computed comparing the number of items in the input factor index and
        the number of items in the output DataFrame index.
        Factor data can be partially dropped due to being flawed itself
        (e.g. NaNs), not having provided enough price data to compute
        forward returns for all factor values, or because it is not possible
        to perform binning.
        Set max_loss=0 to avoid Exceptions suppression.
    zero_aware : bool, optional
        If True, compute quantile buckets separately for positive and negative
        signal values. This is useful if your signal is centered and zero is
        the separation between long and short signals, respectively.
        'quantiles' is None.

    Returns
    -------
    merged_data : pd.DataFrame - MultiIndex
        A MultiIndex Series indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.

        - forward returns column names follow the format accepted by
          pd.Timedelta (e.g. '1D', '30m', '3h15m', '1D1h', etc)

        - 'date' index freq property (merged_data.index.levels[0].freq) is the
          same as that of the input forward returns data. This is currently
          used only in cumulative returns computation
        ::
           -------------------------------------------------------------------
                      |       | 1D  | 5D  | 10D  |factor|group|factor_quantile
           -------------------------------------------------------------------
               date   | asset |     |     |      |      |     |
           -------------------------------------------------------------------
                      | AAPL  | 0.09|-0.01|-0.079|  0.5 |  G1 |      3
                      --------------------------------------------------------
                      | BA    | 0.02| 0.06| 0.020| -1.1 |  G2 |      5
                      --------------------------------------------------------
           2014-01-01 | CMG   | 0.03| 0.09| 0.036|  1.7 |  G2 |      1
                      --------------------------------------------------------
                      | DAL   |-0.02|-0.06|-0.029| -0.1 |  G3 |      5
                      --------------------------------------------------------
                      | LULU  |-0.03| 0.05|-0.009|  2.7 |  G1 |      2
                      --------------------------------------------------------
    """

    initial_amount = float(len(factor.index))

    factor_copy = factor.copy()
    factor_copy.index = factor_copy.index.rename(['date', 'asset'])
    factor_copy = factor_copy[np.isfinite(factor_copy)]

    merged_data = forward_returns.copy()
    merged_data['factor'] = factor_copy

    if groupby is not None:
        if isinstance(groupby, dict):
            diff = set(factor_copy.index.get_level_values(
                'asset')) - set(groupby.keys())
            if len(diff) > 0:
                raise KeyError(
                    "Assets {} not in group mapping".format(
                        list(diff)))

            ss = pd.Series(groupby)
            groupby = pd.Series(index=factor_copy.index,
                                data=ss[factor_copy.index.get_level_values(
                                    'asset')].values)

        if groupby_labels is not None:
            diff = set(groupby.values) - set(groupby_labels.keys())
            if len(diff) > 0:
                raise KeyError(
                    "groups {} not in passed group names".format(
                        list(diff)))

            sn = pd.Series(groupby_labels)
            groupby = pd.Series(index=groupby.index,
                                data=sn[groupby.values].values)

        merged_data['group'] = groupby.astype('category')

    merged_data = merged_data.dropna()

    fwdret_amount = float(len(merged_data.index))

    no_raise = False if max_loss == 0 else True
    quantile_data = quantize_factor(
        merged_data,
        quantiles,
        bins,
        binning_by_group,
        no_raise,
        zero_aware
    )

    merged_data['factor_quantile'] = quantile_data

    merged_data = merged_data.dropna()

    binning_amount = float(len(merged_data.index))

    tot_loss = (initial_amount - binning_amount) / initial_amount
    fwdret_loss = (initial_amount - fwdret_amount) / initial_amount
    bin_loss = tot_loss - fwdret_loss

    # print("Dropped %.1f%% entries from factor data: %.1f%% in forward "
    #       "returns computation and %.1f%% in binning phase "
    #       "(set max_loss=0 to see potentially suppressed Exceptions)." %
    #       (tot_loss * 100, fwdret_loss * 100, bin_loss * 100))

    if tot_loss > max_loss:
        message = ("max_loss (%.1f%%) exceeded %.1f%%, consider increasing it."
                   % (max_loss * 100, tot_loss * 100))
        warnings.warn(message)
        # raise MaxLossExceededError(message)
    else:
        pass
        # print("max_loss is %.1f%%, not exceeded: OK!" % (max_loss * 100))

    return merged_data


def get_clean_factor_and_forward_returns(factor,
                                         prices,
                                         groupby=None,
                                         binning_by_group=False,
                                         quantiles=5,
                                         bins=None,
                                         periods=(1, 5, 10),
                                         filter_zscore=20,
                                         groupby_labels=None,
                                         max_loss=0.35,
                                         zero_aware=False,
                                         cumulative_returns=True,
                                         periods_by_factor=False):
    """
    Formats the factor data, pricing data, and group mappings into a DataFrame
    that contains aligned MultiIndex indices of timestamp and asset. The
    returned data will be formatted to be suitable for Alphalens functions.

    It is safe to skip a call to this function and still make use of Alphalens
    functionalities as long as the factor data conforms to the format returned
    from get_clean_factor_and_forward_returns and documented here

    Parameters
    ----------
    factor : pd.Series - MultiIndex
        A MultiIndex Series indexed by timestamp (level 0) and asset
        (level 1), containing the values for a single alpha factor.
        ::
            -----------------------------------
                date    |    asset   |
            -----------------------------------
                        |   AAPL     |   0.5
                        -----------------------
                        |   BA       |  -1.1
                        -----------------------
            2014-01-01  |   CMG      |   1.7
                        -----------------------
                        |   DAL      |  -0.1
                        -----------------------
                        |   LULU     |   2.7
                        -----------------------

    prices : pd.DataFrame
        A wide form Pandas DataFrame indexed by timestamp with assets
        in the columns.
        Pricing data must span the factor analysis time period plus an
        additional buffer window that is greater than the maximum number
        of expected periods in the forward returns calculations.
        It is important to pass the correct pricing data in depending on
        what time of period your signal was generated so to avoid lookahead
        bias, or  delayed calculations.
        'Prices' must contain at least an entry for each timestamp/asset
        combination in 'factor'. This entry should reflect the buy price
        for the assets and usually it is the next available price after the
        factor is computed but it can also be a later price if the factor is
        meant to be traded later (e.g. if the factor is computed at market
        open but traded 1 hour after market open the price information should
        be 1 hour after market open).
        'Prices' must also contain entries for timestamps following each
        timestamp/asset combination in 'factor', as many more timestamps
        as the maximum value in 'periods'. The asset price after 'period'
        timestamps will be considered the sell price for that asset when
        computing 'period' forward returns.
        ::
            ----------------------------------------------------
                        | AAPL |  BA  |  CMG  |  DAL  |  LULU  |
            ----------------------------------------------------
               Date     |      |      |       |       |        |
            ----------------------------------------------------
            2014-01-01  |605.12| 24.58|  11.72| 54.43 |  37.14 |
            ----------------------------------------------------
            2014-01-02  |604.35| 22.23|  12.21| 52.78 |  33.63 |
            ----------------------------------------------------
            2014-01-03  |607.94| 21.68|  14.36| 53.94 |  29.37 |
            ----------------------------------------------------

    groupby : pd.Series - MultiIndex or dict
        Either A MultiIndex Series indexed by date and asset,
        containing the period wise group codes for each asset, or
        a dict of asset to group mappings. If a dict is passed,
        it is assumed that group mappings are unchanged for the
        entire time period of the passed factor data.
    binning_by_group : bool
        If True, compute quantile buckets separately for each group.
        This is useful when the factor values range vary considerably
        across gorups so that it is wise to make the binning group relative.
        You should probably enable this if the factor is intended
        to be analyzed for a group neutral portfolio
    quantiles : int or sequence[float]
        Number of equal-sized quantile buckets to use in factor bucketing.
        Alternately sequence of quantiles, allowing non-equal-sized buckets
        e.g. [0, .10, .5, .90, 1.] or [.05, .5, .95]
        Only one of 'quantiles' or 'bins' can be not-None
    bins : int or sequence[float]
        Number of equal-width (valuewise) bins to use in factor bucketing.
        Alternately sequence of bin edges allowing for non-uniform bin width
        e.g. [-4, -2, -0.5, 0, 10]
        Chooses the buckets to be evenly spaced according to the values
        themselves. Useful when the factor contains discrete values.
        Only one of 'quantiles' or 'bins' can be not-None
    periods : sequence[int]
        periods to compute forward returns on.
    filter_zscore : int or float, optional
        Sets forward returns greater than X standard deviations
        from the the mean to nan. Set it to 'None' to avoid filtering.
        Caution: this outlier filtering incorporates lookahead bias.
    groupby_labels : dict
        A dictionary keyed by group code with values corresponding
        to the display name for each group.
    max_loss : float, optional
        Maximum percentage (0.00 to 1.00) of factor data dropping allowed,
        computed comparing the number of items in the input factor index and
        the number of items in the output DataFrame index.
        Factor data can be partially dropped due to being flawed itself
        (e.g. NaNs), not having provided enough price data to compute
        forward returns for all factor values, or because it is not possible
        to perform binning.
        Set max_loss=0 to avoid Exceptions suppression.
    zero_aware : bool, optional
        If True, compute quantile buckets separately for positive and negative
        signal values. This is useful if your signal is centered and zero is
        the separation between long and short signals, respectively.
    cumulative_returns : bool, optional
        If True, forward returns columns will contain cumulative returns.
        Setting this to False is useful if you want to analyze how predictive
        a factor is for a single forward day.

    Returns
    -------
    merged_data : pd.DataFrame - MultiIndex
        A MultiIndex Series indexed by date (level 0) and asset (level 1),
        containing the values for a single alpha factor, forward returns for
        each period, the factor quantile/bin that factor value belongs to, and
        (optionally) the group the asset belongs to.
        - forward returns column names follow  the format accepted by
          pd.Timedelta (e.g. '1D', '30m', '3h15m', '1D1h', etc)
        - 'date' index freq property (merged_data.index.levels[0].freq) will be
          set to a trading calendar (pandas DateOffset) inferred from the input
          data (see infer_trading_calendar for more details). This is currently
          used only in cumulative returns computation
        ::
           -------------------------------------------------------------------
                      |       | 1D  | 5D  | 10D  |factor|group|factor_quantile
           -------------------------------------------------------------------
               date   | asset |     |     |      |      |     |
           -------------------------------------------------------------------
                      | AAPL  | 0.09|-0.01|-0.079|  0.5 |  G1 |      3
                      --------------------------------------------------------
                      | BA    | 0.02| 0.06| 0.020| -1.1 |  G2 |      5
                      --------------------------------------------------------
           2014-01-01 | CMG   | 0.03| 0.09| 0.036|  1.7 |  G2 |      1
                      --------------------------------------------------------
                      | DAL   |-0.02|-0.06|-0.029| -0.1 |  G3 |      5
                      --------------------------------------------------------
                      | LULU  |-0.03| 0.05|-0.009|  2.7 |  G1 |      2
                      --------------------------------------------------------

    See Also
    --------
    utils.get_clean_factor
        For use when forward returns are already available.
    """
    forward_returns = compute_forward_returns(
        factor,
        prices,
        periods,
        filter_zscore,
        cumulative_returns,
        periods_by_factor
    )

    factor_data = get_clean_factor(factor, forward_returns, groupby=groupby,
                                   groupby_labels=groupby_labels,
                                   quantiles=quantiles, bins=bins,
                                   binning_by_group=binning_by_group,
                                   max_loss=max_loss, zero_aware=zero_aware)

    return factor_data


def rate_of_return(period_ret, base_period):
    """
    Convert returns to 'one_period_len' rate of returns: that is the value the
    returns would have every 'one_period_len' if they had grown at a steady
    rate

    Parameters
    ----------
    period_ret: pd.DataFrame
        DataFrame containing returns values with column headings representing
        the return period.
    base_period: string
        The base period length used in the conversion
        It must follow pandas.Timedelta constructor format (e.g. '1 days',
        '1D', '30m', '3h', '1D1h', etc)

    Returns
    -------
    pd.DataFrame
        DataFrame in same format as input but with 'one_period_len' rate of
        returns values.
    """
    period_len = period_ret.name
    conversion_factor = (pd.Timedelta(base_period) /
                         pd.Timedelta(period_len))
    return period_ret.add(1).pow(conversion_factor).sub(1)


def std_conversion(period_std, base_period):
    """
    one_period_len standard deviation (or standard error) approximation

    Parameters
    ----------
    period_std: pd.DataFrame
        DataFrame containing standard deviation or standard error values
        with column headings representing the return period.
    base_period: string
        The base period length used in the conversion
        It must follow pandas.Timedelta constructor format (e.g. '1 days',
        '1D', '30m', '3h', '1D1h', etc)

    Returns
    -------
    pd.DataFrame
        DataFrame in same format as input but with one-period
        standard deviation/error values.
    """
    period_len = period_std.name
    conversion_factor = (pd.Timedelta(period_len) /
                         pd.Timedelta(base_period))
    return period_std / np.sqrt(conversion_factor)


def get_forward_returns_columns(columns, require_exact_day_multiple=False):
    """
    Utility that detects and returns the columns that are forward returns
    """

    # If exact day multiples are required in the forward return periods,
    # drop all other columns (e.g. drop 3D12h).
    if require_exact_day_multiple:
        pattern = re.compile(r"^(\d+([D]))+$", re.IGNORECASE)
        valid_columns = [(pattern.match(col) is not None) for col in columns]

        if sum(valid_columns) < len(valid_columns):
            warnings.warn(
                "Skipping return periods that aren't exact multiples"
                + " of days."
            )
    else:
        pattern = re.compile(r"^(\d+([Dhms]|ms|us|ns]))+$", re.IGNORECASE)
        valid_columns = [(pattern.match(col) is not None) for col in columns]

    return columns[valid_columns]


def timedelta_to_string(timedelta):
    """
    Utility that converts a pandas.Timedelta to a string representation
    compatible with pandas.Timedelta constructor format

    Parameters
    ----------
    timedelta: pd.Timedelta

    Returns
    -------
    string
        string representation of 'timedelta'
    """
    c = timedelta.components
    format = ''
    if c.days != 0:
        format += '%dD' % c.days
    if c.hours > 0:
        format += '%dh' % c.hours
    if c.minutes > 0:
        format += '%dm' % c.minutes
    if c.seconds > 0:
        format += '%ds' % c.seconds
    if c.milliseconds > 0:
        format += '%dms' % c.milliseconds
    if c.microseconds > 0:
        format += '%dus' % c.microseconds
    if c.nanoseconds > 0:
        format += '%dns' % c.nanoseconds
    return format


def timedelta_strings_to_integers(sequence):
    """
    Converts pandas string representations of timedeltas into integers of days.

    Parameters
    ----------
    sequence : iterable
        List or array of timedelta string representations, e.g. ['1D', '5D'].

    Returns
    -------
    sequence : list
        Integer days corresponding to the input sequence, e.g. [1, 5].
    """
    return list(map(lambda x: pd.Timedelta(x).days, sequence))


def add_custom_calendar_timedelta(input, timedelta, freq):
    """
    Add timedelta to 'input' taking into consideration custom frequency, which
    is used to deal with custom calendars, such as a trading calendar

    Parameters
    ----------
    input : pd.DatetimeIndex or pd.Timestamp
    timedelta : pd.Timedelta
    freq : pd.DataOffset (CustomBusinessDay, Day or BusinessDay)

    Returns
    -------
    pd.DatetimeIndex or pd.Timestamp
        input + timedelta
    """
    if not isinstance(freq, (Day, BusinessDay, CustomBusinessDay)):
        raise ValueError("freq must be Day, BDay or CustomBusinessDay")
    days = timedelta.components.days
    offset = timedelta - pd.Timedelta(days=days)
    return input + freq * days + offset


def diff_custom_calendar_timedeltas(start, end, freq):
    """
    Compute the difference between two pd.Timedelta taking into consideration
    custom frequency, which is used to deal with custom calendars, such as a
    trading calendar

    Parameters
    ----------
    start : pd.Timestamp
    end : pd.Timestamp
    freq : CustomBusinessDay (see infer_trading_calendar)
    freq : pd.DataOffset (CustomBusinessDay, Day or BDay)

    Returns
    -------
    pd.Timedelta
        end - start
    """
    if not isinstance(freq, (Day, BusinessDay, CustomBusinessDay)):
        raise ValueError("freq must be Day, BusinessDay or CustomBusinessDay")

    weekmask = getattr(freq, 'weekmask', None)
    holidays = getattr(freq, 'holidays', None)

    if weekmask is None and holidays is None:
        if isinstance(freq, Day):
            weekmask = 'Mon Tue Wed Thu Fri Sat Sun'
            holidays = []
        elif isinstance(freq, BusinessDay):
            weekmask = 'Mon Tue Wed Thu Fri'
            holidays = []

    if weekmask is not None and holidays is not None:
        # we prefer this method as it is faster
        actual_days = np.busday_count(np.array(start).astype('datetime64[D]'),
                                      np.array(end).astype('datetime64[D]'),
                                      weekmask, holidays)
    else:
        # default, it is slow
        actual_days = pd.date_range(start, end, freq=freq).shape[0] - 1
        if not freq.onOffset(start):
            actual_days -= 1

    timediff = end - start
    delta_days = timediff.components.days - actual_days
    return timediff - pd.Timedelta(days=delta_days)



import torch 
import torch.nn.functional as F
from abc import ABC, abstractmethod
import torch.nn as nn

import pandas as pd
import numpy as np
# from ..stockcalendar import CALENDAR_TOOL
from torch.utils.data import TensorDataset, DataLoader, Dataset,ConcatDataset
import time
import os
import copy
from tqdm import tqdm




class DLmodel(ABC):
    """
    深度学习模型的抽象类, 子类要实现必要的方法.
    
    基类功能:
    - self.__init__():
        -- self.check: 检查GPU资源可用性
        -- self.class_model: DL模型架构
        -- self._model_params(dict): DL模型参数设定
        -- self._dtype: 模型精度，默认float
        -- self.set_params: 设定模型参数，每次调用都会重置模型
    
    - self.check: 检查GPU资源可用性
    - self.set_params: 设定模型参数，每次调用都会重置模型
    - self.save: 存储模型
    
    子类功能:
    - self.train(): 子类必须实现模型的训练功能   
    - self.test(): 子类必须实现模型的测试功能 
    
    """
    
    def __init__(self,model_params:dict,class_model,dtype='float',device_num=0) -> None:
        self.check()
        self._state = []
        self.class_model = class_model
        self._model_params:dict = model_params
        self._dtype = dtype
        self.set_params()
    
    def check(self):
        print('CUDA is available:{}'.format(torch.cuda.is_available()))
        print('GPU device count:{}'.format(torch.cuda.device_count()))
        num = torch.cuda.device_count()
        if not torch.cuda.is_available():
            print('Use CPU!!!')
        else:
            for i in range(num):
                print('GPU device name {}:{}'.format(i,torch.cuda.get_device_name(i)))
        if self.device_num >= num:
            raise ValueError('You device is not in the available list!')
            
        self.device = torch.device("cuda:{}".format(self.device_num) if torch.cuda.is_available() else "cpu")
        print("Use the GPU {}!!!!!".format(self.device_num))
        
    def set_params(self):
        if self._dtype == 'float':
            self.model = self.class_model(**self._model_params).float().to(self.device)
        else:
            self.model = self.class_model(**self._model_params).to(self.device)
    
    def r2_score(self,target:torch.tensor,output:torch.tensor):
        # 计算总平方和 (SStot)
        mean_target = torch.mean(target,dim=0)
        total_sum_squares = torch.sum((target - mean_target) ** 2,dim=0)

        # 计算残差平方和 (SSres)
        residual_sum_squares = torch.sum((target - output) ** 2,dim=0)

        # 计算R² score
        r2score = 1 - (residual_sum_squares / total_sum_squares)
        return r2score
    
    def precision(self,y_true:torch.tensor,y_pred:torch.tensor,num:int=None):
        if num is None:
            num = y_pred.max() + 1
        y_true = y_true.view(-1)
        y_pred = y_pred.view(-1)
        
        result = []
        for i in range(num):
            TP = ((y_true == i) & (y_pred == i)).sum().to(torch.float32)  # True Positive
            FP = ((y_true != i) & (y_pred == i)).sum().to(torch.float32)  # False Positive
            FN = ((y_true == i) & (y_pred != i)).sum().to(torch.float32)  # False Negative

            precision = TP / (TP + FP + 1e-6)  # 防止分母为0
            result.append(precision.item())
        return result
    
    def recall(self,y_true:torch.tensor,y_pred:torch.tensor,num:int=None):
        if num is None:
            num = y_pred.max() + 1
        y_true = y_true.view(-1)
        y_pred = y_pred.view(-1)
        
        result = []
        for i in range(num):
            TP = ((y_true == i) & (y_pred == i)).sum().to(torch.float32)  # True Positive
            FN = ((y_true == i) & (y_pred != i)).sum().to(torch.float32)  # False Negative

            recall = TP / (TP + FN + 1e-6)  # 防止分母为0
            result.append(recall.item())
        return result
    
    def accuracy(self,y_true, y_pred):
        y_pred = y_pred.view(-1)
        y_true = y_true.view(-1)

        correct_predictions = torch.sum(y_true == y_pred).item()

        accuracy = correct_predictions / y_true.size(0)
        return accuracy

    def f1_score_macro(self,y_true,y_pred, num:int=None):
        if num is None:
            num = y_pred.max() + 1
        y_true = y_true.view(-1)
        y_pred = y_pred.view(-1)
        
        f1_scores = []
        for i in range(num):
            TP = ((y_true == i) & (y_pred == i)).sum().to(torch.float32)  # True Positive
            FP = ((y_true != i) & (y_pred == i)).sum().to(torch.float32)  # False Positive
            FN = ((y_true == i) & (y_pred != i)).sum().to(torch.float32)  # False Negative

            precision = TP / (TP + FP + 1e-6)  # 防止分母为0
            recall = TP / (TP + FN + 1e-6)  # 防止分母为0
            f1 = 2 * (precision * recall) / (precision + recall + 1e-6)  # 防止分母为0
            f1_scores.append(f1)

        f1_score_macro = torch.mean(torch.tensor(f1_scores)).item()
        return f1_score_macro
    
    def reload_model(self,path):
        self.model = torch.load(path)
            
    
    @abstractmethod
    def train(self):
        raise NotImplementedError("should implement in the derived class")
    
    @abstractmethod
    def test(self):
        raise NotImplementedError("should implement in the derived class")
    
    
    def save(self,path):
        self.model.eval()
        torch.save(self.model.to(self.device),path)
        
    def save_state(self):
        self._state.append(copy.deepcopy(self.model.state_dict()))
    
    def reset_state(self,state):
        self.model.load_state_dict(state, strict=True)
    
    def ave_state(self,window=3,skip=1):
        if len(self._state) == 0:
            raise ValueError("None state in self._state, ensure you use the function to save them after train.")
        if len(self.ave_state) < window:
            print("The length of self._state < window, the average will be restricted to the existing state list!")
        averaged_params = {}
        used_params = self._state[-window:].copy()
        used_params.reverse()
        used_params = used_params[::skip]
        # 遍历所有保存的参数
        for param_name in used_params[0].keys():
            # 初始化参数总和
            param_sum = torch.zeros_like(used_params[0][param_name])
            
            # 累加每个周期的参数
            for params in used_params:
                param_sum += params[param_name]
            
            # 计算平均值
            averaged_params[param_name] = param_sum / len(used_params)
        return averaged_params


class ClassifyModelFrame(DLmodel):
    """
    Introduce:
        基于分类DL模型的训练和测试框架
    
    Function:
    - __init__:提供你的模型设计和模型参数设定，会自动帮你生成和存储模型，继承DLmodel方法
    - train: 基础的训练框架，请根据你的需要输出一些你的监控指标和不同的批次生成方式
    - test: 提供你的测试集数据，输出预测结果，防止内存不足提供根据一定的批量分批预测的方式。
    """
    def __init__(self, model_params: dict, class_model, dtype='float') -> None:
        super().__init__(model_params, class_model, dtype)
    
    def train(self,dataset,EPOCHS,batch_size=512,shuffle=True,lr=0.001,gap=[-0.005,0.005]):
        gap = torch.tensor(gap).to(self.device)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
        loss_func = nn.CrossEntropyLoss()
        optim = torch.optim.Adam(self.model.parameters(),lr=lr)
        
        for epoch in range(EPOCHS):
            self.model.train()

            for batch_idx, (data,target) in enumerate(dataloader):
                data = data.float().to(self.device)
                target = target.float().to(self.device)
                label = target.reshape(-1)
                label = torch.bucketize(label,gap)
                
                optim.zero_grad()
                output = self.model(data)
                
                l = loss_func(output,label)

                l.backward()
                optim.step()
                
                if batch_idx % 50 == 0:
                    print("train epoch:{}, batch_idx:{}".format(epoch,batch_idx))
            
            
            print("FINISH: train epoch:{}".format(epoch))
            print("")

        return 1
    
    
    def test(self,data:torch.tensor,model,batch_size=1024):
        model = model.to(self.device)
        result = []
        with torch.no_grad():
            model.eval()
            for i in range(0,data.shape[0],batch_size):
                feature = data[i:i+batch_size].float().to(self.device)
                label = F.softmax(model(feature),dim=1).cpu().detach().numpy()
                result.append(label)
        result = np.vstack(result)
        return result
    
    
    
class RegresionModelFrame(DLmodel):
    
    """
    Introduce:
        基于回归DL模型的训练和测试框架
    
    Function:
    - __init__:提供你的模型设计和模型参数设定，会自动帮你生成和存储模型，继承DLmodel方法
    - train: 提供你的训练数据集，批次大小，训练轮次，早停条件，学习率等，回归训练以R2score达标作为早停条件
    - test: 提供你的测试集数据，和同样顺序的index
    """
    
    def __init__(self, model_params, class_model,dtype='float') -> None:
        super().__init__(model_params, class_model, dtype)
        
    def train(self,dataset,EPOCHS,batch_size=512,scale=1,shuffle=True,loss='Huber', lr=0.001):
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)
        
        if loss == 'Huber':
            loss_func = nn.HuberLoss()
        elif loss == 'L1':
            loss_func = nn.L1Loss()
        elif loss == 'MSE': 
            loss_func = nn.MSELoss()
        else:
            loss_func = nn.MSELoss()

        
        optim = torch.optim.Adam(self.model.parameters(),lr=lr)

        for epoch in range(EPOCHS):
            self.model.train()
            for batch_idx, (data,target) in enumerate(dataloader):
                
                data = data.float().to(self.device)
                target = target.float().to(self.device)*scale
                optim.zero_grad()
                output = self.model(data)
                
                l = loss_func(output,target)
                
                l.backward()
                
                optim.step()
                
                if batch_idx % 50 == 0:
                    print("train epoch:{}, batch_idx:{}".format(epoch+1,batch_idx))
            
            print("FINISH: train epoch:{}".format(epoch))
            print("")
        
        return 1
    
    def test(self,data:torch.tensor,model,batch_size=2048):
        model = model.cuda()
        result = []
        with torch.no_grad():
            model.eval()
            for i in range(0,data.shape[0],batch_size):
                feature = data[i:i+batch_size].float().to(self.device)
                label = model(feature).cpu().detach().numpy()
                result.append(label)
        result = np.vstack(result)
        return result
    



class StockDataset(Dataset):
    """
    时序dataset, 截面特征set num_feature_times = 1, then 处理 dimension in the neural network
    """

    def __init__(self, features, label=None, num_feature_times=1, is_train=True, feature_stan=None,industry=None):
        """
        feature: index=(date,feature), columns=stock
        label: index = date, columns=stock
        num_feature_times: 时序长度
        is_train: 是否为训练集
        """
        self.dates = features.index.get_level_values(0).unique()
        self.features = features
        self.label = label
        self.num_feature_times = num_feature_times
        self.is_train = is_train
        self.num_features = len(self.features.loc[self.dates[0]])
        self.feature_stan = feature_stan
        if industry is None:
            self._industry_matrix = None
        else:
            self._industry_matrix = self.industry_matrix(industry)

    def __len__(self):
        return len(self.dates) - self.num_feature_times + 1

    def convert_to_3d_numpy(self, df, label=None):
        df: pd.DataFrame = df.dropna(axis=1)  ## index = (date,features), columns=stock
        if self.is_train:
            if label is None:
                raise ValueError("when you set is_train==True, you should provide label")
            label = label.dropna()
            
            assets_feature = set(df.columns)
            assets_label = set(label.index)
            assets = list(assets_feature & assets_label)
            assets.sort()
            
            df = df[assets]
            label = label.loc[assets]
        else:
            assets = list(df.columns)
        
        # 获取形状信息
        num_stocks = len(assets)
        np_array = df.to_numpy()
        
        if self._industry_matrix is not None:
            r_matrix = self._industry_matrix.loc[assets].values
            r_matrix = r_matrix @ r_matrix.T
            comm_features = r_matrix @ np_array.T
            comm_features = comm_features.reshape(
                num_stocks, self.num_feature_times, self.num_features
            )

        # 重塑数组为[N, T, F]
        features = np_array.T.reshape(
            num_stocks, self.num_feature_times, self.num_features
        )
        if self._industry_matrix is not None:
            ### 若有行业, 则拼接为[N,T,F+F]
            features = np.concatenate([features,comm_features],axis=2)
        if self.feature_stan == "panel":
            features = self.panel_stan(features) ## 轴1不动，沿着特征，将time*stock矩阵做标准化，并用0值填充空缺值
        elif self.feature_stan == "two-step":
            features = self.two_step_stan(features) ## 先对时序除以平均值，再沿股票做标准
        
            
             

        if self.is_train:
            labels = label.values
            return assets, features, labels
        else:
            return assets, features

    def panel_stan(self, matrix):
        mean = np.nanmean(matrix, axis=(0, 1), keepdims=True)
        std = np.nanstd(matrix, axis=(0, 1), keepdims=True)
        return np.nan_to_num((matrix - mean) / std)

    def two_step_stan(self, matrix):
        # divided by timeseries mean and then cross-sectional zscore
        ts_mean = np.nanmean(matrix, axis=1, keepdims=True)
        matrix /= ts_mean
        cs_mean = np.nanmean(matrix, axis=0, keepdims=True)
        cs_std = np.nanstd(matrix, axis=0, keepdims=True)
        return np.nan_to_num((matrix - cs_mean) / cs_std)
    
    def industry_matrix(self,industry:pd.Series):
        """
        根据index为asset,value为industry的Series生成01的DataFrame(index=asset,columns=industry)
        """
        industry.index.name = 'asset'
        industry.name = 'industry'
        industry:pd.DataFrame = industry.reset_index()
        industry['is_true'] = 1
        industry = industry.pivot(index='asset',columns='industry',values='is_true').fillna(0)
        return industry

    def __getitem__(self, idx):
        start_date = self.dates[idx]
        end_date = self.dates[idx + self.num_feature_times - 1]

        if self.is_train:
            label = self.label.loc[end_date].dropna()
            stocks, features, labels = self.convert_to_3d_numpy(
                self.features.loc[start_date:end_date],label
            )
            return end_date, stocks, features, labels
        else:
            stocks, features = self.convert_to_3d_numpy(
                self.features.loc[start_date:end_date],label=None
            )
            return end_date, stocks, features

    @staticmethod
    def data_wash(factor):
        """
        factor: pd.DataFrame. index=(date,asset)
        """
        factor = factor.dropna(how='all')
        factor = factor.swaplevel().sort_index()
        factor = factor.groupby(level=0).ffill()
        factor = factor.unstack(level=1).T
        factor = factor.swaplevel().sort_index()
        return factor

class StockDataLoader(DataLoader):
    """
    custom stock market dataloader, 每个batch只包含单日全部股票
    """

    def __init__(
        self,
        df,
        label=None,
        is_train=True,
        feature_stan=None,
        num_feature_times=1,
        batch_size=1,
        shuffle=False,
        num_workers=0,
    ):
        dataset = StockDataset(
            df,
            label,
            num_feature_times=num_feature_times,
            is_train=is_train,
            feature_stan=feature_stan,
        )

        self.is_train = is_train

        super().__init__(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            collate_fn=self.collate_fn,
        )

    def collate_fn(self, batch):
        if self.is_train:
            date, stocks, features, labels = zip(*batch)
            features = torch.tensor(features[0], dtype=torch.float)
            labels = torch.tensor(labels[0], dtype=torch.float)
            return date[0], stocks[0], features, labels
        else:
            date, stocks, features = zip(*batch)
            features = torch.tensor(features[0], dtype=torch.float)
            return date[0], stocks[0], features


if __name__ == "__main__":
    pass


def gen_train_dataloader(df,label,dataset_path,num_feature_times,batch_size,shuffle=True,hard_stop=50):
    """维护一个不大于hard_stop GB 的数据集文件夹，以数据生成点的日期命名文件，大于hard_stop强制停止工作，df和label格式与StockDataset格式相同"""
    def bytes_to_GB(num):
        return round(num/1073741824,3)

    def file_size(path):
        file_lst = os.listdir(path)
        total = 0
        for name in file_lst:
            total += os.path.getsize(path+'/' + name)
        return bytes_to_GB(total)
    print("Gen train dataloader.......")
    t1 = time.time()
    stock_dataset = StockDataset(df,label,is_train=True,num_feature_times=num_feature_times)
        
    ### 删除无用数据集
    exist_dataset = os.listdir(dataset_path)
    use_dataset = list(map(lambda x: 'dataset.train.'+x.strftime("%Y%m%d"),stock_dataset.dates[num_feature_times-1:]))
    for name in exist_dataset:
        if name not in use_dataset:
            os.remove(dataset_path+'/'+name)
    
    ### 生成使用的数据集
    exist_dataset = os.listdir(dataset_path)
    print("生成：")
    for idx in tqdm(range(len(stock_dataset))):
        ed =  stock_dataset.dates[idx+num_feature_times-1]
        ed = ed.strftime('%Y%m%d')
        if 'dataset.train.'+ed in exist_dataset:
            continue
        
        _, _, features, labels = stock_dataset.__getitem__(idx)
        features = torch.tensor(features).float()
        labels = torch.tensor(labels).float()
        dataset = TensorDataset(features,labels)
        torch.save(dataset,dataset_path + '/' + 'dataset.train.{}'.format(ed))
        test_size = file_size(dataset_path)
        if test_size > hard_stop:
            raise ValueError("file size is {}GB > hard_stop {}GB, Stop it!".format(test_size,hard_stop))
    print("    dataset_path file size is {} GB".format(file_size(dataset_path)))           
    ### 生成训练用数据集和dataloader
    exist_dataset = os.listdir(dataset_path)
    datasets = []
    print("合并：")
    for name in tqdm(exist_dataset):
        datasets.append(torch.load(dataset_path+'/'+name))
    train_dataset = ConcatDataset(datasets)
    dataloader = DataLoader(train_dataset,batch_size=batch_size,shuffle=shuffle,num_workers=12)
    t2 = time.time()
    print("    gen time: {}".format(round(t2-t1,2)))
    return dataloader
    
    
    











# trade_dates = CALENDAR_TOOL.trade_date_in_range(start_date='20010101',end_date='20261201').copy(deep=True)

# class ByTimeDataLoader:
#     """
#     每回遍历不重复的从data_set中跳出一个数据点
#     """
#     def __init__(self, data_set, data_range=None):
#         """
#         初始化迭代器
#         :param data_set: 数据集，列表形式
#         :param data_range: 生成随机数据的范围，默认为data_set的列表长度
#         """
#         self.data_set = data_set
#         if data_range == None:
#             self.data_range = len(data_set)
#         else:
#             self.data_range = data_range
#         self.current_batch = 0
#         self.random_shuffle = [_ for _ in range(self.data_range)]
#         random.shuffle(self.random_shuffle)
    
#     def reset(self):
#         random.shuffle(self.random_shuffle)
#         self.current_batch = 0

#     def __iter__(self):
#         """返回迭代器对象本身"""
#         return self

#     def __next__(self):
#         """生成下一个随机批量的数据"""
#         if self.current_batch < self.data_range:
#             batch_data = self.data_set[self.random_shuffle[self.current_batch]]
#             self.current_batch += 1
#             return batch_data
#         else:
#             self.reset()
#             raise StopIteration  # 如果所有批次都已经生成，则停止迭代    

# class GenDataSet:
#     """
#     用于生成各种类型数据集的类函数
#     """
    
#     @staticmethod
#     def base_train_dataset(factor,price,dates,seq_len,forward,dropmid=None,dtype='float',assets=None):
#         """
#         Params:
#         -factor: pd.DataFrame. index=pd.MultiIndex, level0 = asset, level1 = date;
#         -price: pd.DataFrame. index=date, columns = asset
#         -dates: list(datetime). 提供你要生成数据点的时间点，将逐次生成训练数据，可通过控制这个参数防止偷看未来数据
#         -seq_len: int. 获取的因子矩阵的时序长度
#         -forward: int. 获取的未来收益率长度
#         -dropmid=None: float. 在进行标签筛选时是否扔掉中间的部分数据
#         -dtype='float': str. 生成的数据集数据类型
        
#         Return:
#         (index,dataset): 与dataset相对应的index, 方便直接使用训练集输出测试预测因子
        
#         Attention: 
#         - 建议你提供的数据相对于时间点来说要足够长, 否则会丢弃很多长度不够的数据点。
#         - 对于时间点前seqlen长度的数据，若空缺比例超过0.1会丢弃该数据点，所以请在因子输入时dropna
#         """
#         seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
#         if assets is None:
#             assets = list(set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
#         else:
#             assets = list(set(list(assets)) & set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
#         price = price[assets]
        
        
#         def single_point(asset,date,seq_len):
#             pre_date = seek_backward_dates.loc[date]
#             df = factor.loc[asset].loc[pre_date:date]
            
#             if len(df) < seq_len*0.9:
#                 return None
#             else:
#                 df = factor.loc[asset].loc[:date].iloc[-seq_len:].values
#                 if len(df) != seq_len:
#                     return None
#                 return ((asset,date),df)
        
#         X_ = []
#         y_ = []
        
#         index = []
        
#         def drop_mid_Series(sr:Series,dropmid):
#             sr = sr.sort_values()
#             l = len(sr)
#             if l < 100:
#                 return sr
#             return pd.concat([sr.iloc[:int(l*(1/2-dropmid/2))],sr.iloc[int(l*(1/2+dropmid/2)):]])
        
#         forward_returns = (price.shift(-forward)/price - 1)
#         for date in dates:
#             y:Series = forward_returns.loc[date].dropna()
#             if len(y) == 0:
#                 continue
#             if dropmid is not None:
#                 y = drop_mid_Series(y,dropmid)
#             for asset in y.index:
#                 dv = single_point(asset,date,seq_len)
#                 if dv is None:
#                     continue
                
#                 X_.append(dv[1])
#                 y_.append(y.loc[asset])
#                 index.append(dv[0])

#         if dtype == 'float':
#             X_ = torch.tensor(np.array(X_)).float()
#             y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1).float()
#         else:
#             X_ = torch.tensor(np.array(X_))
#             y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1)
            
#         index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
#         return (index,TensorDataset(X_,y_))
    
    
    # @staticmethod
    # def market_train_dataset(factor,price,market,dates,seq_len,forward,dropmid=None,dtype='float',droptop=None):
    #     """
    #     Params:
    #     -factor: pd.DataFrame. index=pd.MultiIndex, level0 = asset, level1 = date;
    #     -price: pd.DataFrame. index=date, columns = asset
    #     -market: pd.Series. index=date, 你需要额外提供一个市场数据(收盘价)，标签将会是超市场收益
    #     -dates: list(datetime). 提供你要生成数据点的时间点，将逐次生成训练数据，可通过控制这个参数防止偷看未来数据
    #     -seq_len: int. 获取的因子矩阵的时序长度
    #     -forward: int. 获取的未来收益率长度
    #     -dropmid=None: float. 在进行标签筛选时是否扔掉中间的部分数据
    #     -dtype='float': str. 生成的数据集数据类型
        
    #     Return:
    #     (index,dataset): 与dataset相对应的index, 方便直接使用训练集输出测试预测因子
    #     """
    #     seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
    #     def single_point(asset,date,seq_len):
    #         pre_date = seek_backward_dates.loc[date]
    #         df = factor.loc[asset].loc[pre_date:date]
    #         if len(df) < seq_len*0.9:
    #             return None
    #         else:
    #             df = factor.loc[asset].loc[:date].iloc[-seq_len:].values
    #             if len(df) != seq_len:
    #                 return None
    #             return ((asset,date),df)
        
    #     X_ = []
    #     y_ = []
        
    #     index = []
        
    #     def drop_mid_Series(sr:Series,dropmid):
    #         sr = sr.sort_values()
    #         l = len(sr)
    #         if l < 100:
    #             return sr
    #         return pd.concat([sr.iloc[:int(l*(1/2-dropmid/2))],sr.iloc[int(l*(1/2+dropmid/2)):]])
        
    #     def drop_top_Series(sr:Series,droptop):
    #         sr = sr.sort_values()
    #         l = len(sr)
    #         if l < 100:
    #             return sr
    #         return sr.iloc[:int(l*(1-droptop))]
        
    #     forward_returns:DataFrame = (price.shift(-forward)/price - 1)
    #     forward_market = (market.shift(-forward)/market - 1)
    #     forward_exceed = forward_returns.subtract(forward_market,axis=0)
    #     for date in dates:
    #         y:Series = forward_exceed.loc[date].dropna()
    #         if len(y) == 0:
    #             continue
    #         if dropmid is not None:
    #             y = drop_mid_Series(y,dropmid)
    #         if droptop is not None:
    #             y = drop_top_Series(y,droptop)
    #         for asset in y.index:
    #             dv = single_point(asset,date,seq_len)
    #             if dv is None:
    #                 continue
                
    #             X_.append(dv[1])
    #             y_.append(y.loc[asset])
    #             index.append(dv[0])

    #     if dtype == 'float':
    #         X_ = torch.tensor(np.array(X_)).float()
    #         y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1).float()
    #     else:
    #         X_ = torch.tensor(np.array(X_))
    #         y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1)
            
    #     index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
    #     return (index,TensorDataset(X_,y_))
    
    
    # @staticmethod
    # def seq_train_dataset(factor,price,dates,seq_len,forward,assets=None,dtype='float'):
    #     """
    #     Params:
    #     -factor: pd.DataFrame. index=pd.MultiIndex, level0 = asset, level1 = date;
    #     -price: pd.DataFrame. index=date, columns = asset
    #     -dates: list(datetime). 提供你要生成数据点的时间点，将逐次生成训练数据，可通过控制这个参数防止偷看未来数据
    #     -seq_len: int. 获取的因子矩阵的时序长度
    #     -forward: int. 获取的未来收益率长度
    #     -dtype='float': str. 生成的数据集数据类型
        
    #     Return:
    #     (index,dataset): 你得到的标签将会是forward长度的收益率序列，而非单个收益率.与dataset相对应的index, 方便直接使用训练集输出测试预测因子.
    #     """
    #     if assets is None:
    #         assets = list(set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     else:
    #         assets = list(set(list(assets)) & set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     price = price[assets]
        
        
    #     seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
    #     def single_point(asset,date,seq_len):
    #         pre_date = seek_backward_dates.loc[date]
    #         df = factor.loc[asset].loc[pre_date:date]
    #         if len(df) < seq_len*0.9:
    #             return None
    #         else:
    #             df = factor.loc[asset].loc[:date].iloc[-seq_len:].values
    #             if len(df) != seq_len:
    #                 return None
    #             return ((asset,date),df)
        
    #     X_ = []
    #     y_ = []
        
    #     index = []
    #     returns = price/price.shift(1)-1
    #     forward_returns = (price.shift(-forward)/price - 1)
    #     for date in dates:
    #         y = returns.loc[date:].iloc[1:forward+1].T.dropna()
    #         if len(y) == 0:
    #             continue
    #         for asset in y.index:
    #             dv = single_point(asset,date,seq_len)
    #             if dv is None:
    #                 continue
                
    #             X_.append(dv[1])
    #             y_.append(y.loc[asset].values)
    #             index.append(dv[0])

    #     if dtype == 'float':
    #         X_ = torch.tensor(np.array(X_)).float()
    #         y_ = torch.tensor(np.array(y_)).float()
    #     else:
    #         X_ = torch.tensor(np.array(X_))
    #         y_ = torch.tensor(np.array(y_))
            
    #     index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
    #     return (index,TensorDataset(X_,y_))
    
    # @staticmethod
    # def Seq_train_dataset(factor,price,dates,seq_len,forward:int=7,dtype='float',assets=None):
    #     """
    #     Introduce:
    #     获得标签为未来forward日的数据序列的数据集，要保证factor已经dropna，空缺日期比例超过10%则丢弃该数据点
        
    #     Params:
    #     -factor: pd.DataFrame. index=pd.MultiIndex, level0 = asset, level1 = date;
    #     -price: pd.DataFrame. index=date, columns = asset
    #     -dates: list(datetime). 提供你要生成数据点的时间点，将逐次生成训练数据，可通过控制这个参数防止偷看未来数据
    #     -seq_len: int. 获取的因子矩阵的时序长度
    #     -forward: int. 获取的未来收益率长度, 输出的标签为未来forward天的收益率序列，可以构造夏普比等特殊标签
    #     -dropmid=None: float. 在进行标签筛选时是否扔掉中间的部分数据
    #     -dtype='float': str. 
    #     -assets=None: list. 提供你筛选后的股票，如果没有则默认为price表提供的股票
        
    #     Return:
    #     (index,dataset): 与dataset相对应的index, 方便直接使用训练集输出测试预测因子
        
    #     Attention: 
    #     - 建议你提供的数据相对于时间点来说要足够长, 否则会丢弃很多长度不够的数据点。
    #     - 对于时间点前seqlen长度的数据，若空缺比例超过0.1会丢弃该数据点，所以请在因子输入时dropna
    #     """
    #     seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
    #     if assets is None:
    #         assets = list(set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     else:
    #         assets = list(set(list(assets)) & set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     price:DataFrame = price[assets]
    #     returns = price.pct_change()
        
    #     def single_point(asset,date,seq_len):
    #         pre_date = seek_backward_dates.loc[date]
            
    #         df = factor.loc[asset].loc[:date]
    #         test_df = df.loc[pre_date:]
    #         if len(test_df) < seq_len*0.9:
    #             return None
    #         else:
    #             df = df.iloc[-seq_len:].values
    #             if np.isnan(df).sum() > 0:
    #                 return None
    #             if len(df) != seq_len:
    #                 return None
    #             return ((asset,date),df)
        
    #     X_ = []
    #     label = []

    #     index = []
    #     forward_returns = []
    #     for i in range(forward):
    #         forward_returns.append(returns.shift(-i-1).stack(dropna=False))
    #     forward_returns = pd.concat(forward_returns,axis=1).sort_index()
    #     for date in dates:
    #         y_ = forward_returns.loc[date].dropna()
    #         if len(y_) == 0:
    #             continue
    #         for asset in y_.index:
    #             dv = single_point(asset,date,seq_len)
    #             if dv is None:
    #                 continue
                
    #             X_.append(dv[1])
    #             label.append(y_.loc[asset].values)
    #             index.append(dv[0])

    #     if dtype == 'float':
    #         X_ = torch.tensor(np.array(X_)).float()
    #         label = torch.tensor(np.array(label)).float()
    #     else:
    #         X_ = torch.tensor(np.array(X_)).float()
    #         label = torch.tensor(np.array(label)).float()
            
    #     index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
    #     return (index,TensorDataset(X_,label))
    
    
    # def None_train_dataset(factor,price,dates,seq_len,forward,none_num=None,dropmid=None,dtype='float',assets=None):
    #     """
    #     Introduce:
    #     主要添加了空缺值的处理逻辑，同时尽量不损失数据，空缺比例高于10%则丢弃，否则用ffill填补空缺值.
        
    #     Params:
    #     -factor: pd.DataFrame. index=pd.MultiIndex, level0 = asset, level1 = date;
    #     -price: pd.DataFrame. index=date, columns = asset
    #     -dates: list(datetime). 提供你要生成数据点的时间点，将逐次生成训练数据，可通过控制这个参数防止偷看未来数据
    #     -seq_len: int. 获取的因子矩阵的时序长度
    #     -forward: int. 获取的未来收益率长度
    #     -dropmid=None: float. 在进行标签筛选时是否扔掉中间的部分数据
    #     -dtype='float': str. 
    #     -assets=None: list. 提供你筛选后的股票，如果没有则默认为price表提供的股票
    #     -none_num: Serier. 提供因子每一行是否有空缺值的bool值Series，不提供会内部计算，有点慢.
        
    #     Return:
    #     (index,dataset): 与dataset相对应的index, 方便直接使用训练集输出测试预测因子
        
    #     Attention: 
    #     - 建议你提供的数据相对于时间点来说要足够长, 否则会丢弃很多长度不够的数据点。
    #     """
    #     seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
    #     if assets is None:
    #         assets = list(set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     else:
    #         assets = list(set(list(assets)) & set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     price = price[assets]
    #     if none_num is None:
    #         none_num = factor.isna().sum(axis=1)>0
        
        
    #     def single_point(asset,date,seq_len):
    #         pre_date = seek_backward_dates.loc[date]
    #         n = none_num.loc[asset].loc[pre_date:date]
            
    #         if n.sum() > seq_len*0.1:
    #             return None
    #         else:
    #             df = factor.loc[asset].loc[:date].ffill().iloc[-seq_len:].values
    #             if np.isnan(df).sum() > 0:
    #                 return None
    #             if len(df) != seq_len:
    #                 return None
    #             return ((asset,date),df)
        
    #     X_ = []
    #     y_ = []
        
    #     index = []
        
    #     def drop_mid_Series(sr:Series,dropmid):
    #         sr = sr.sort_values()
    #         l = len(sr)
    #         if l < 100:
    #             return sr
    #         return pd.concat([sr.iloc[:int(l*(1/2-dropmid/2))],sr.iloc[int(l*(1/2+dropmid/2)):]])
        
    #     forward_returns = (price.shift(-forward)/price - 1)
    #     for date in dates:
    #         y:Series = forward_returns.loc[date].dropna()
    #         if len(y) == 0:
    #             continue
    #         if dropmid is not None:
    #             y = drop_mid_Series(y,dropmid)
    #         for asset in y.index:
    #             dv = single_point(asset,date,seq_len)
    #             if dv is None:
    #                 continue
                
    #             X_.append(dv[1])
    #             y_.append(y.loc[asset])
    #             index.append(dv[0])

    #     if dtype == 'float':
    #         X_ = torch.tensor(np.array(X_)).float()
    #         y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1).float()
    #     else:
    #         X_ = torch.tensor(np.array(X_))
    #         y_ = torch.tensor(np.array(y_)).reshape(len(y_),-1)
            
    #     index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
    #     return (index,TensorDataset(X_,y_))
    
    
    # @staticmethod
    # def Multi_None_train_dataset(factor,price,dates,seq_len,forward=(1,3,5,10),none_num=None,dtype='float',assets=None):
    #     """
    #     Introduce:
    #     主要添加了空缺值的处理逻辑，同时尽量不损失数据，空缺比例高于10%则丢弃，否则用ffill填补空缺值. 得到的标签则是多个未来区间的收益率，用于不同长度的预测。
        
    #     Params:
    #     -factor: pd.DataFrame. index=pd.MultiIndex, level0 = asset, level1 = date;
    #     -price: pd.DataFrame. index=date, columns = asset
    #     -dates: list(datetime). 提供你要生成数据点的时间点，将逐次生成训练数据，可通过控制这个参数防止偷看未来数据
    #     -seq_len: int. 获取的因子矩阵的时序长度
    #     -forward: tuple. 获取的未来收益率长度, 可接受多个时间按长度，最后输出的标签每一列都是不同forward的数据
    #     -dropmid=None: float. 在进行标签筛选时是否扔掉中间的部分数据
    #     -dtype='float': str. 
    #     -assets=None: list. 提供你筛选后的股票，如果没有则默认为price表提供的股票
    #     -none_num: Serier. index=pd.MultiIndex, level0 = asset, level1 = date. 提供因子每一行是否有空缺值的bool值Series，不提供会内部计算，有点慢.
        
    #     Return:
    #     (index,dataset): 与dataset相对应的index, 方便直接使用训练集输出测试预测因子
        
    #     Attention: 
    #     - 建议你提供的数据相对于时间点来说要足够长, 否则会丢弃很多长度不够的数据点。
    #     """
    #     seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
    #     if assets is None:
    #         assets = list(set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     else:
    #         assets = list(set(list(assets)) & set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     price = price[assets]
    #     if none_num is None:
    #         none_num = factor.isna().sum(axis=1)>0
        
        
    #     def single_point(asset,date,seq_len):
    #         pre_date = seek_backward_dates.loc[date]
    #         n = none_num.loc[asset].loc[pre_date:date]
            
    #         if n.sum() > seq_len*0.1:
    #             return None
    #         else:
    #             df = factor.loc[asset].loc[:date].ffill().iloc[-seq_len:].values
    #             if np.isnan(df).sum() > 0:
    #                 return None
    #             if len(df) != seq_len:
    #                 return None
    #             return ((asset,date),df)
        
    #     X_ = []
    #     label = []

    #     index = []
        
    #     forward_returns = []
    #     for fw in forward:
    #         forward_returns.append((price.shift(-fw)/price - 1))
        
    #     for date in dates:
    #         y_ = []
    #         for i in range(len(forward_returns)):
    #             y_.append(forward_returns[i].loc[date])
    #         y_ = pd.concat(y_,axis=1).dropna()
    #         if len(y_) == 0:
    #             continue
    #         for asset in y_.index:
    #             dv = single_point(asset,date,seq_len)
    #             if dv is None:
    #                 continue
                
    #             X_.append(dv[1])
    #             label.append(y_.loc[asset].values)
    #             index.append(dv[0])

    #     if dtype == 'float':
    #         X_ = torch.tensor(np.array(X_)).float()
    #         label = torch.tensor(np.array(label)).float()
    #     else:
    #         X_ = torch.tensor(np.array(X_)).float()
    #         label = torch.tensor(np.array(label)).float()
            
    #     index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
    #     return (index,TensorDataset(X_,label))
    
    # @staticmethod
    # def Seq_None_train_dataset(factor,price,dates,seq_len,forward:int=7,none_num=None,dtype='float',assets=None):
    #     """
    #     Introduce:
    #     主要添加了空缺值的处理逻辑，同时尽量不损失数据，空缺比例高于10%则丢弃，否则用ffill填补空缺值. 得到的标签是未来forward日的收益率序列，用于更加复杂的标签构造分析。
        
    #     Params:
    #     -factor: pd.DataFrame. index=pd.MultiIndex, level0 = asset, level1 = date;
    #     -price: pd.DataFrame. index=date, columns = asset
    #     -dates: list(datetime). 提供你要生成数据点的时间点，将逐次生成训练数据，可通过控制这个参数防止偷看未来数据
    #     -seq_len: int. 获取的因子矩阵的时序长度
    #     -forward: int. 获取的未来收益率长度, 输出的标签为未来forward天的收益率序列，可以构造夏普比等特殊标签
    #     -dropmid=None: float. 在进行标签筛选时是否扔掉中间的部分数据
    #     -dtype='float': str. 
    #     -assets=None: list. 提供你筛选后的股票，如果没有则默认为price表提供的股票
    #     -none_num: Serier. index=pd.MultiIndex, level0 = asset, level1 = date. 提供因子每一行是否有空缺值的bool值Series，不提供会内部计算，有点慢.
        
    #     Return:
    #     (index,dataset): 与dataset相对应的index, 方便直接使用训练集输出测试预测因子
        
    #     Attention: 
    #     - 建议你提供的数据相对于时间点来说要足够长, 否则会丢弃很多长度不够的数据点。
    #     - 对于时间点前seqlen长度的数据，若空缺比例超过0.1会丢弃该数据点，所以请在因子输入时dropna
    #     """
    #     seek_backward_dates = Series(trade_dates,index=trade_dates).shift(seq_len).dropna()
    #     if assets is None:
    #         assets = list(set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     else:
    #         assets = list(set(list(assets)) & set(list(factor.index.get_level_values(0))) & set(list(price.columns)))
    #     price:DataFrame = price[assets]
    #     returns = price.pct_change()
    #     if none_num is None:
    #         none_num = factor.isna().sum(axis=1)>0
        
        
    #     def single_point(asset,date,seq_len):
    #         pre_date = seek_backward_dates.loc[date]
    #         n = none_num.loc[asset].loc[pre_date:date]
            
    #         if n.sum() > seq_len*0.1:
    #             return None
    #         else:
    #             df = factor.loc[asset].loc[:date].ffill().iloc[-seq_len:].values
    #             if np.isnan(df).sum() > 0:
    #                 return None
    #             if len(df) != seq_len:
    #                 return None
    #             return ((asset,date),df)
        
    #     X_ = []
    #     label = []

    #     index = []
    #     forward_returns = []
    #     for i in range(forward):
    #         forward_returns.append(returns.shift(-i-1).stack(dropna=False))
    #     forward_returns = pd.concat(forward_returns,axis=1).sort_index()
    #     for date in dates:
    #         y_ = forward_returns.loc[date].dropna()
    #         if len(y_) == 0:
    #             continue
    #         for asset in y_.index:
    #             dv = single_point(asset,date,seq_len)
    #             if dv is None:
    #                 continue
                
    #             X_.append(dv[1])
    #             label.append(y_.loc[asset].values)
    #             index.append(dv[0])

    #     if dtype == 'float':
    #         X_ = torch.tensor(np.array(X_)).float()
    #         label = torch.tensor(np.array(label)).float()
    #     else:
    #         X_ = torch.tensor(np.array(X_)).float()
    #         label = torch.tensor(np.array(label)).float()
            
    #     index = pd.MultiIndex.from_tuples(index,names=['asset','date'])
        
    #     return (index,TensorDataset(X_,label))


import os
import warnings

import alphalens as al
from factor import *

warnings.filterwarnings("ignore")


# 以下都是实现的因子Demo

class Alpha005(Factor):
    # 因子的名称， 不能与基础因子冲突。
    name = 'alpha005'
    # 获取数据的最长时间窗口，返回的是日级别的数据。
    max_window = 10
    # 依赖的基础因子名称，包含 'open', 'close', 'low', 'high', 'volume', 'vwap', 'returns' 字段
    dependencies = ['open', 'close', 'vwap']

    def calc(self, data):
        return rank((data['open'] - (ts_sum(data['vwap'], 10) / 10))) * (-1 * abs(rank((data['close'] - data['vwap']))))


class ALPHA006(Factor):
    # 因子的名称， 不能与基础因子冲突。
    name = 'alpha006'
    # 获取数据的最长时间窗口，返回的是日级别的数据。
    max_window = 10
    # 依赖的基础因子名称，包含 'open', 'close', 'low', 'high', 'volume', 'vwap', 'returns' 字段
    dependencies = ['open', 'volume']

    def calc(self, data):
        df = -1 * correlation(data['open'], data['volume'], self.max_window)
        return df.replace([-np.inf, np.inf], 0).fillna(value=0)


class SMA5(Factor):
    # 因子的名称， 不能与基础因子冲突。
    name = 'sma5'
    # 获取数据的最长时间窗口，返回的是日级别的数据。
    max_window = 5
    # 依赖的基础因子名称，包含 'open', 'close', 'low', 'high', 'volume', 'vwap', 'returns' 字段
    dependencies = ['close']

    def calc(self, data):
        return data['close'].rolling(5).mean()


class EMA5(Factor):
    # 因子的名称， 不能与基础因子冲突。
    name = 'ema5'
    # 获取数据的最长时间窗口，返回的是日级别的数据。
    max_window = 5
    # 依赖的基础因子名称，包含 'open', 'close', 'low', 'high', 'volume', 'vwap', 'returns' 字段
    dependencies = ['close']

    def calc(self, data):
        return data['close'].ewm(com=9, span=5, min_periods=self.max_window).mean()


class WMA5(Factor):
    # 因子的名称， 不能与基础因子冲突。
    name = 'wma5'
    # 获取数据的最长时间窗口，返回的是日级别的数据。
    max_window = 5
    # 依赖的基础因子名称，包含 'open', 'close', 'low', 'high', 'volume', 'vwap', 'returns' 字段
    dependencies = ['close']

    def calc(self, data):
        return data['close'].rolling(self.max_window).apply(
            lambda x: x[::-1].cumsum().sum() * 2 / self.max_window / (self.max_window + 1))


if __name__ == '__main__':
    # 测试开始时间, 测试结束时间
    # start_date = datetime.datetime(2022, 8, 15)
    # end_date = datetime.datetime(2022, 9, 25)

    print(os.environ)

    start_date = 20211015
    end_date = 20220925

    # 沪深300股票池
    stocks = sd.get_index_stocks("000300")

    print(stocks)

    # 自定义因子
    factor_data = calc_factors(securities=stocks, factors=[ALPHA006()], start_date=start_date,
                               end_date=end_date)
    # factor_data = calc_factors(securities=stocks, factors=[Alpha005(), ALPHA006()], start_date=start_date,
    #                            end_date=end_date)

    # 需要转换成pd.MultiIndex结构
    print("*" * 200)
    print(factor_data)

    # multiIndexFactors = factor_data['alpha006']
    multiIndexFactors = factor_data

    print("*" * 200)
    print(multiIndexFactors)

    forwardPeriodPrices = sd.get_price(stocks, start_date=start_date, end_date=end_date,
                                       fields=['close'])

    forwardPeriodPrices = forwardPeriodPrices.pivot(index='date', columns='issue',
                                                    values='close')

    print("*" * 200)
    print(forwardPeriodPrices)

    # periods是计算未来1、5、10、15的收益率
    data = al.utils.get_clean_factor_and_forward_returns(factor=multiIndexFactors, prices=forwardPeriodPrices,
                                                         quantiles=5, periods=(1, 5, 10, 15), max_loss=1)

    print("*" * 200)
    print(data)

    sd_get_Analysis_results = al.tears.sd_get_Analysis_results(data, longshort=True, plot=True)

    print("*" * 200)
    print(sd_get_Analysis_results)

    sd_get_return_results = al.tears.sd_get_return_results(data,turnover_Data=None, longshort=True, plot=True)

    print("*" * 200)
    print(sd_get_return_results)
    # al.tears.create_full_tear_sheet(data)



from .utils import get_clean_factor_and_forward_returns
from ..datatools import DataHandle
from ..backtest import QuickBackTestor
from pandas import Series,DataFrame
import pandas as pd
import numpy as np
from ..data_statistics import StatTtest
from .utils import MaxLossExceededError



def factor_neutral_groupby(factor:DataFrame,group:dict,names=['date','group']):
    '''按组和日期对因子数据进行标准化，需要的时间较长，记得跑完存盘'''
    factor = factor.copy(deep=True)
    factor['group'] = factor.index.get_level_values(level=1).map(group)
    return factor.groupby(names).transform(DataHandle.normalize)

class SingleFactorAnalysis:

    def __init__(self,start_date,end_date,factor:Series,prices:DataFrame,periods=(1,5,20),quantiles=5,groupby=None,periods_by_factor=True, is_limit_buy=None, is_limit_sell=None):
        self.start_date = start_date
        self.end_date = end_date
        self._prices = prices.loc[start_date:end_date]
        self._factor = factor.loc[start_date:end_date]
        self.universe = prices.columns
        self.n_assets = len(self.universe)
        self._group_by = groupby
        # self._periods = periods
        self._quantiles = quantiles
        self.periods_by_factor = periods_by_factor
        self.backtest = QuickBackTestor(self.start_date,self.end_date,self._prices,is_limit_buy=is_limit_buy,is_limit_sell=is_limit_sell)
        self._cleaned_factor = get_clean_factor_and_forward_returns(factor=self._factor,
                                                                    prices=self._prices,
                                                                    quantiles=quantiles,
                                                                    periods=periods,
                                                                    groupby=groupby,
                                                                    filter_zscore=None,
                                                                    zero_aware=False,
                                                                    periods_by_factor=periods_by_factor)
        

    def IC(self, look_forward=None, method="spearman"):
        if self.periods_by_factor:
            look_forward = '1T'
        else:
            if look_forward is None:
                raise ValueError('look_forward should not be None if you do not set periods_by_factor.')
            look_forward = f"{look_forward}D"
        return self._cleaned_factor.groupby("date").apply(lambda x: x[look_forward].corr(x["factor"], method=method))
    
    def cumIC(self,look_forward=None, method="spearman"):
        return self.IC(look_forward,method).cumsum()    
    
    def get_group_assets(self,factor_quantile):
        tmp = self._cleaned_factor[self._cleaned_factor['factor_quantile'] == factor_quantile]["factor_quantile"]
        return tmp
    
    def get_group_positions(self, factor_quantile: int):
        tmp = self._cleaned_factor[self._cleaned_factor['factor_quantile'] == factor_quantile]["factor_quantile"].unstack().reindex(self.universe, axis=1)
        weights = (tmp > 0) / np.repeat((tmp>0).sum(axis=1).values.reshape(-1,1), self.n_assets, axis=1)
        return weights
    
    def get_long_short_positions(self):
        weights1 = self.get_group_positions(factor_quantile=1)
        weights2 = self.get_group_positions(factor_quantile=self._quantiles)
        return weights2/2-weights1/2
    
    def group_return(self,factor_quantile:int,is_limit=True,fee=0):
        weights = self.get_group_positions(factor_quantile)

        self.backtest.run_backtest(weights,is_limit)
        return self.backtest.fee_adjust_rewards(fee)
    
    def groups_return(self,is_limit=True,fee=0):
        """这里是returns,不是PnL,懒得改"""
        PnLs = []
        for i in range(1,self._quantiles + 1):

            PnL = self.group_return(factor_quantile=i,is_limit=is_limit,fee=fee)
            PnLs.append(PnL)
        PnLs = pd.concat(PnLs,axis=1)
        PnLs.columns = [f'group{i}' for i in range(1,self._quantiles + 1)]
        return PnLs
    
    def long_short_return(self,is_limit=True):
        weights = self.get_long_short_positions()
        return self.backtest.run_backtest(weights,is_limit)
    
    def top_return(self,num,is_limit=True,fee=0):
        nf = self._factor.unstack().sort_index()
        weights = nf.apply(lambda x: x.nlargest(num), axis=1)
        weights[~weights.isna()] = 1/num
        weights = weights.T.reindex(self.universe).T
        weights = weights.fillna(0)
        self.backtest.run_backtest(weights,is_limit)
        return self.backtest.fee_adjust_rewards(fee=fee)
    
    def hypo_testing(self,look_forward=None,method='RLM',alpha=0.05):
        if self.periods_by_factor:
            look_forward = '1T'
        else:
            if look_forward is None:
                raise ValueError('look_forward should not be None if you do not set periods_by_factor.')
            look_forward = f"{look_forward}D"
        test_bool = []
        for date in self._cleaned_factor.index.get_level_values(0).drop_duplicates():
            X = self._cleaned_factor.loc[date]['factor']
            y = self._cleaned_factor.loc[date][look_forward]
            test = StatTtest(X,y,method=method)
            test_bool.append(test.t_test(alpha=alpha))
        test_frame = DataFrame(np.array(test_bool))
        if len(test_frame.columns) != 3:
            print(test_frame) 
            print(self._cleaned_factor)
        test_frame.columns = ['beta','u','d']
        total = len(test_frame.index)
        if total == 0:
            return Series(np.zeros(8),
                          index = ['正向显著比例','负向显著比例','正-负','同向显著次数比','状态切换次数比','同向-切换','因子方向','总检验次数'])
        u = test_frame['u'].sum()
        d = test_frame['d'].sum()
        pass_frame = test_frame[(test_frame['u']==True)|(test_frame['d']==True)]
        sep = pd.concat([pass_frame['beta'],pass_frame['u'].shift(1)],axis=1).dropna()
        sep.columns = ['0','1']
        same = (sep['0'] == sep['1']).sum()
        dif = (sep['0'] != sep['1']).sum()
        if u>=d:
            direct = 1
        else:
            direct = -1
        return Series([u/total,d/total,(u-d)/total, same/total,dif/total,(same-dif)/total,direct,total],
                      index = ['正向显著比例','负向显著比例','正-负','同向显著次数比','状态切换次数比','同向-切换','因子方向','总检验次数'])       


# class MultiFactorAnalysis:
#     """
#     This class allow you to do multi-factor analysis, you can input factor DataFrame. But if your file is too big, it could be slow.

#     Params:
#     -start_date: str|datetime.
#     -end_date: str|datetime.
#     -factor:DataFrame. MultiIndex with (date,assets), columns should be factor name.
#     -prices:DataFrame. index should be datetime, columns should be assets name(same as factor assets).
#     -periods:Any=(1,5,20). the time periods you calculate the forward returns, it follows the factor frequency! The cleand factor frame -columns will be '1D','5D','20D'. 
#     -quantiles:int=5. the group number you seperate the factor every date.
#     -groupby:dict=None. keys are assets name, values are group name, usually industry map.
#     -periods_by_factor:bool=True. True, then the periods will be related with factor frequency, and periods input will be useless.

#     Attention:
#         None
#     """
#     def __init__(self,start_date,end_date,factor:DataFrame,prices:DataFrame,periods=(1,5,20),quantiles=5,groupby=None,periods_by_factor=True):
#         self.start_date = start_date
#         self.end_date = end_date
#         self._prices = prices.loc[start_date:end_date]
#         self._factor = factor.loc[start_date:end_date]
#         self.universe = prices.columns
#         self.n_assets = len(self.universe)
#         self._group_by = groupby
#         self._periods = periods
#         self._quantiles = quantiles
#         self.periods_by_factor = periods_by_factor
#         self._factor_name = self._factor.columns
#         self.backtest = QuickBackTestor(self.start_date,self.end_date,self._prices)
#         cleaned_factors = {}
#         wrong_fac = []
#         for fac in self._factor_name:
#             try:
#                 cleaned_factors[fac] = get_clean_factor_and_forward_returns(factor=self._factor[fac],
#                                                                         prices=self._prices,
#                                                                         periods=periods,
#                                                                         quantiles=quantiles,
#                                                                         groupby=groupby,
#                                                                         filter_zscore=None,
#                                                                         zero_aware=False,
#                                                                         periods_by_factor=periods_by_factor)
#             except MaxLossExceededError:
#                 wrong_fac.append(fac)
#                 continue
#         print('You have wrong factor which max_loss exceed 35%, these have been deleted, please check.')
#         print(wrong_fac)
#         self._factor_name = list(set(self._factor_name).difference(wrong_fac))
#         self._wrong_factor = wrong_fac
#         self._cleaned_factors = cleaned_factors
#         del self._factor

    
#     def multi_ICs(self, fac_name:list=None,look_forward=None, method="spearman"):
#         if fac_name is None:
#             fac_name = self._factor_name

#         if self.periods_by_factor:
#             look_forward = '1T'
#         else:
#             if look_forward is None:
#                 raise ValueError('look_forward should not be None if you do not set periods_by_factor.')
#             look_forward = f"{look_forward}D"

#         IC_lst = []
#         for fac in fac_name:
#             cf = self._cleaned_factors[fac]
#             IC_lst.append(cf.groupby("date").apply(lambda x: x[look_forward].corr(x["factor"], method=method)))
#         ICs = pd.concat(IC_lst,axis=1)
#         ICs.columns = fac_name
#         return ICs
    
#     def multi_cumICs(self, fac_name:list=None,look_forward=None, method="spearman"):
#         return self.multi_ICs(fac_name,look_forward,method).cumsum()
    
#     def single_get_group_assets(self,fac:str,factor_quantile):
#         cf = self._cleaned_factors[fac]
#         tmp = cf[cf['factor_quantile'] == factor_quantile]["factor_quantile"]
#         return tmp
    
#     def single_get_group_positions(self,fac:str,factor_quantile: int):
#         """某年因子的某组的等权组合权重"""
#         cf = self._cleaned_factors[fac]
#         tmp = cf[cf['factor_quantile'] == factor_quantile]["factor_quantile"].unstack().reindex(self.universe, axis=1)
#         weights = (tmp > 0) / np.repeat((tmp>0).sum(axis=1).values.reshape(-1,1), self.n_assets, axis=1)
#         return weights
    
#     def single_get_long_short_positions(self,fac:str):
#         """某年因子的某组的多空组合权重"""
#         weights1 = self.single_get_group_positions(fac,factor_quantile=1)
#         weights2 = self.single_get_group_positions(fac,factor_quantile=self._quantiles)
#         return weights2/2-weights1/2
    
#     def single_group_return(self,fac:str,factor_quantile:int,limit_up=True,limit_down=True):
#         weights = self.get_group_positions(fac,factor_quantile)
#         return self.backtest.run_backtest(weights,limit_up,limit_down)
    
#     def single_groups_return(self,fac:str,limit_up=True,limit_down=True):
#         """这里是returns,不是PnL,懒得改"""
#         returns = []
#         for i in range(1,self._quantiles + 1):
#             rt = self.single_group_return(fac,factor_quantile=i,limit_up=limit_up,limit_down=limit_down)
#             returns.append(rt)
#         returns = pd.concat(returns,axis=1)
#         returns.columns = [f'group{i}' for i in range(1,self._quantiles + 1)]
#         return returns
    
#     def multi_long_short_return(self,fac_name:list=None,limit_up=True,limit_down=True):
#         returns = []
#         if fac_name is None:
#             fac_name = self._factor_name
#         for fac in fac_name:
#             weights = self.single_get_long_short_positions(fac)
#             rt = self.backtest.run_backtest(weights,limit_up,limit_down)
#             returns.append(rt)
#         returns = pd.concat(returns,axis=1)
#         returns.columns = fac_name
#         return returns
    
#     def single_hypo_testing(self,fac:str,look_forward=None,method='RLM',alpha=0.05):
#         if self.periods_by_factor:
#             look_forward = '1T'
#         else:
#             if look_forward is None:
#                 raise ValueError('look_forward should not be None if you do not set periods_by_factor.')
#             look_forward = f"{look_forward}D"
#         test_bool = []
#         cf = self._cleaned_factors[fac]
#         for date in cf.index.get_level_values(0).drop_duplicates():
#             X = cf.loc[date]['factor']
#             y = cf.loc[date][look_forward]
#             test = StatTtest(X,y,method=method)
#             test_bool.append(test.t_test(alpha=alpha))
#         test_frame = DataFrame(np.array(test_bool))
#         if len(test_frame.columns) != 3:
#             print(test_frame) 
#             print(cf)
#         test_frame.columns = ['beta','u','d']
#         total = len(test_frame.index)
#         if total == 0:
#             return Series(np.zeros(8),
#                           index = ['正向显著比例','负向显著比例','正-负','同向显著次数比','状态切换次数比','同向-切换','因子方向','总检验次数'])
#         u = test_frame['u'].sum()
#         d = test_frame['d'].sum()
#         pass_frame = test_frame[(test_frame['u']==True)|(test_frame['d']==True)]
#         sep = pd.concat([pass_frame['beta'],pass_frame['u'].shift(1)],axis=1).dropna()
#         sep.columns = ['0','1']
#         same = (sep['0'] == sep['1']).sum()
#         dif = (sep['0'] != sep['1']).sum()
#         if u>=d:
#             direct = 1
#         else:
#             direct = -1
#         return Series([u/total,d/total,(u-d)/total, same/total,dif/total,(same-dif)/total,direct,total],
#                       index = ['正向显著比例','负向显著比例','正-负','同向显著次数比','状态切换次数比','同向-切换','因子方向','总检验次数']) 

#     def multi_hypo_testing(self,fac_name:list=None,look_forward=None,method='RLM',alpha=0.05):
#         if fac_name is None:
#             fac_name = self._factor_name
#         results = []
#         for fac in fac_name:
#             ht = self.single_hypo_testing(fac,look_forward,method,alpha)
#             results.append(ht)
#         results = pd.concat(results,axis=1)
#         results.columns = fac_name
#         return results



class MultiFactorAnalysis:

    def __init__(self,start_date,end_date,factor:Series,prices:DataFrame,periods=(1,5,20),quantiles=5,groupby=None,periods_by_factor=True, is_limit_buy=None, is_limit_sell=None):
        self.start_date = start_date
        self.end_date = end_date
        self._prices = prices.loc[start_date:end_date]
        self._factor = factor.loc[start_date:end_date]
        self.universe = prices.columns
        self.n_assets = len(self.universe)
        self._groupby = groupby
        self._periods = periods
        self.fac = None
        self.result = {}
        self._quantiles = quantiles
        self.periods_by_factor = periods_by_factor
        self.backtest = QuickBackTestor(self.start_date,self.end_date,self._prices)
        self.backtest.gen_is_limit(is_limit_buy,is_limit_sell)

    
    def set_fac(self,fac:str):
        self.fac = fac
        
    def clean(self):
        del self.result
        self.result = {}
    
    def gen_cleaned_factor(self):
        if self.fac is None:
            print('Please set factor name first!')
            return None
        if self.fac in self.result.keys():
            self._cleaned_factor = self.result[self.fac]
            print('Already Existed.')
            return None
        
        factor:DataFrame = self._factor[self.fac].unstack().sort_index()
        num = factor.subtract(factor.mean(axis=1),axis=0)
        den = factor.std(axis=1)
        factor = num.divide(den,axis=0).stack().sort_index()
        self._cleaned_factor = get_clean_factor_and_forward_returns(factor=factor,
                                                                    prices=self._prices,
                                                                    quantiles=self._quantiles,
                                                                    periods=self._periods,
                                                                    groupby=self._groupby,
                                                                    filter_zscore=None,
                                                                    zero_aware=False,
                                                                    periods_by_factor=self.periods_by_factor)
        self.result[self.fac] = self._cleaned_factor
        print("Generate Complete.")
        return None

        
        

    def IC(self, look_forward=None, method="spearman"):
        if self.periods_by_factor:
            look_forward = '1T'
        else:
            if look_forward is None:
                raise ValueError('look_forward should not be None if you do not set periods_by_factor.')
            look_forward = f"{look_forward}D"
        return self._cleaned_factor.groupby("date").apply(lambda x: x[look_forward].corr(x["factor"], method=method))
    
    def cumIC(self,look_forward=None, method="spearman"):
        return self.IC(look_forward,method).cumsum()    
    
    def get_group_assets(self,factor_quantile):
        tmp = self._cleaned_factor[self._cleaned_factor['factor_quantile'] == factor_quantile]["factor_quantile"]
        return tmp
    
    def get_group_positions(self, factor_quantile: int):
        tmp = self._cleaned_factor[self._cleaned_factor['factor_quantile'] == factor_quantile]["factor_quantile"].unstack().reindex(self.universe, axis=1)
        weights = (tmp > 0) / np.repeat((tmp>0).sum(axis=1).values.reshape(-1,1), self.n_assets, axis=1)
        return weights
    
    def get_long_short_positions(self):
        weights1 = self.get_group_positions(factor_quantile=1)
        weights2 = self.get_group_positions(factor_quantile=self._quantiles)
        return weights2/2-weights1/2
    
    def group_return(self,factor_quantile:int,is_limit=True):
        weights = self.get_group_positions(factor_quantile)

        return self.backtest.run_backtest(weights,is_limit)
    
    def groups_return(self,is_limit=True):
        """这里是returns,不是PnL,懒得改"""
        PnLs = []
        for i in range(1,self._quantiles + 1):

            PnL = self.group_return(factor_quantile=i,is_limit=is_limit)
            PnLs.append(PnL)
        PnLs = pd.concat(PnLs,axis=1)
        PnLs.columns = [f'group{i}' for i in range(1,self._quantiles + 1)]
        return PnLs
    
    def long_short_return(self,is_limit=True):
        weights = self.get_long_short_positions()
        return self.backtest.run_backtest(weights,is_limit)
    
    def top_return(self,num,is_limit=True):
        nf = self._factor.unstack().sort_index()
        weights = nf.apply(lambda x: x.nlargest(num), axis=1)
        weights[~weights.isna()] = 1/num
        weights = weights.T.reindex(self.universe).T
        weights = weights.fillna(0)
        return self.backtest.run_backtest(weights,is_limit)
    
    def hypo_testing(self,look_forward=None,method='RLM',alpha=0.05):
        if self.periods_by_factor:
            look_forward = '1T'
        else:
            if look_forward is None:
                raise ValueError('look_forward should not be None if you do not set periods_by_factor.')
            look_forward = f"{look_forward}D"
        test_bool = []
        for date in self._cleaned_factor.index.get_level_values(0).drop_duplicates():
            X = self._cleaned_factor.loc[date]['factor']
            y = self._cleaned_factor.loc[date][look_forward]
            test = StatTtest(X,y,method=method)
            test_bool.append(test.t_test(alpha=alpha))
        test_frame = DataFrame(np.array(test_bool))
        if len(test_frame.columns) != 3:
            print(test_frame) 
            print(self._cleaned_factor)
        test_frame.columns = ['beta','u','d']
        total = len(test_frame.index)
        if total == 0:
            return Series(np.zeros(8),
                          index = ['正向显著比例','负向显著比例','正-负','同向显著次数比','状态切换次数比','同向-切换','因子方向','总检验次数'])
        u = test_frame['u'].sum()
        d = test_frame['d'].sum()
        pass_frame = test_frame[(test_frame['u']==True)|(test_frame['d']==True)]
        sep = pd.concat([pass_frame['beta'],pass_frame['u'].shift(1)],axis=1).dropna()
        sep.columns = ['0','1']
        same = (sep['0'] == sep['1']).sum()
        dif = (sep['0'] != sep['1']).sum()
        if u>=d:
            direct = 1
        else:
            direct = -1
        return Series([u/total,d/total,(u-d)/total, same/total,dif/total,(same-dif)/total,direct,total],
                      index = ['正向显著比例','负向显著比例','正-负','同向显著次数比','状态切换次数比','同向-切换','因子方向','总检验次数'])

    



from . import tears

def create_summary_tear_sheet(
        factor_data, long_short=True, group_neutral=False
):
    return tears.create_summary_tear_sheet(
        factor_data, long_short, group_neutral)

def create_information_tear_sheet(
        factor_data, group_neutral=False, by_group=False
):
    return tears.create_information_tear_sheet(
        factor_data, group_neutral, by_group)

def create_turnover_tear_sheet(factor_data, turnover_periods=None):
    return tears.create_turnover_tear_sheet(factor_data, turnover_periods)


def create_full_tear_sheet(factor_data,
                           long_short=True,
                           group_neutral=False,
                           by_group=False):
    return tears.create_full_tear_sheet(factor_data,
                           long_short,
                           group_neutral,
                           by_group)

def create_event_returns_tear_sheet(factor_data,
                                    returns,
                                    avgretplot=(5, 15),
                                    long_short=True,
                                    group_neutral=False,
                                    std_bar=True,
                                    by_group=False):
    return tears.create_event_returns_tear_sheet(factor_data,
                                    returns,
                                    avgretplot,
                                    long_short,
                                    group_neutral,
                                    std_bar,
                                    by_group)

def create_event_study_tear_sheet(factor_data,
                                  returns,
                                  avgretplot=(5, 15),
                                  rate_of_ret=True,
                                  n_bars=50):
    return tears.create_event_study_tear_sheet(factor_data,
                                  returns,
                                  avgretplot,
                                  rate_of_ret,
                                  n_bars)

def sd_plot_factor_statistics_table(factor_data):
    return tears.sd_plot_factor_statistics_table(factor_data)


def sd_get_return_results(factor_data, turnover_Data, ana_mode, plot):
    return tears.sd_get_return_results(factor_data, turnover_Data, ana_mode, plot)


def sd_get_information_result(factor_data, plot):
    return tears.sd_get_information_result(factor_data, plot)


def sd_factor_information_coefficient(factor_data, window):
    return tears.sd_factor_information_coefficient(factor_data, window)

def sd_get_turnover_result(factor_data, plot):
    return tears.sd_get_turnover_result(factor_data, plot)


def sd_get_Analysis_results(factor_data, ana_mode='longshort', plot=False):
    return tears.sd_get_Analysis_results(factor_data, ana_mode, plot)


__all__ = []
__all__.extend(['create_summary_tear_sheet','create_information_tear_sheet','create_turnover_tear_sheet','create_full_tear_sheet',
                'create_event_returns_tear_sheet','create_event_study_tear_sheet','sd_plot_factor_statistics_table','sd_get_return_results',
                'sd_get_information_result','sd_factor_information_coefficient','sd_get_turnover_result','sd_get_Analysis_results'])


#
# Copyright 2017 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
from unittest import TestCase
from parameterized import parameterized
from numpy import nan
from pandas import (DataFrame, date_range, Timedelta, concat)

from .. tears import (create_returns_tear_sheet,
                      create_information_tear_sheet,
                      create_turnover_tear_sheet,
                      create_summary_tear_sheet,
                      create_full_tear_sheet,
                      create_event_returns_tear_sheet,
                      create_event_study_tear_sheet)

from .. utils import get_clean_factor_and_forward_returns


class TearsTestCase(TestCase):

    tickers = ['A', 'B', 'C', 'D', 'E', 'F']

    factor_groups = {'A': 1, 'B': 2, 'C': 1, 'D': 2, 'E': 1, 'F': 2}

    price_data = [[1.25**i, 1.50**i, 1.00**i, 0.50**i, 1.50**i, 1.00**i]
                  for i in range(1, 51)]

    factor_data = [[3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                   [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                   [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                   [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                   [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                   [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2],
                   [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, 4, 2, 1, nan, nan], [3, 4, 2, 1, nan, nan],
                   [3, nan, nan, 1, 4, 2], [3, nan, nan, 1, 4, 2]]

    event_data = [[1, nan, nan, nan, nan, nan],
                  [4, nan, nan, 7, nan, nan],
                  [nan, nan, nan, nan, nan, nan],
                  [nan, 3, nan, 2, nan, nan],
                  [1, nan, nan, nan, nan, nan],
                  [nan, nan, 2, nan, nan, nan],
                  [nan, nan, nan, 2, nan, nan],
                  [nan, nan, nan, 1, nan, nan],
                  [2, nan, nan, nan, nan, nan],
                  [nan, nan, nan, nan, 5, nan],
                  [nan, nan, nan, 2, nan, nan],
                  [nan, nan, nan, nan, nan, nan],
                  [2, nan, nan, nan, nan, nan],
                  [nan, nan, nan, nan, nan, 5],
                  [nan, nan, nan, 1, nan, nan],
                  [nan, nan, nan, nan, 4, nan],
                  [5, nan, nan, 4, nan, nan],
                  [nan, nan, nan, 3, nan, nan],
                  [nan, nan, nan, 4, nan, nan],
                  [nan, nan, 2, nan, nan, nan],
                  [5, nan, nan, nan, nan, nan],
                  [nan, 1, nan, nan, nan, nan],
                  [nan, nan, nan, nan, 4, nan],
                  [0, nan, nan, nan, nan, nan],
                  [nan, 5, nan, nan, nan, 4],
                  [nan, nan, nan, nan, nan, nan],
                  [nan, nan, 5, nan, nan, 3],
                  [nan, nan, 1, 2, 3, nan],
                  [nan, nan, nan, 5, nan, nan],
                  [nan, nan, 1, nan, 3, nan]]

    #
    # business days calendar
    #
    bprice_index = date_range(start='2015-1-10', end='2015-3-22', freq='B')
    bprice_index.name = 'date'
    bprices = DataFrame(index=bprice_index, columns=tickers, data=price_data)

    bfactor_index = date_range(start='2015-1-15', end='2015-2-25', freq='B')
    bfactor_index.name = 'date'
    bfactor = DataFrame(index=bfactor_index, columns=tickers,
                        data=factor_data).stack()

    #
    # full calendar
    #
    price_index = date_range(start='2015-1-10', end='2015-2-28')
    price_index.name = 'date'
    prices = DataFrame(index=price_index, columns=tickers, data=price_data)

    factor_index = date_range(start='2015-1-15', end='2015-2-13')
    factor_index.name = 'date'
    factor = DataFrame(index=factor_index, columns=tickers,
                       data=factor_data).stack()

    #
    # intraday factor
    #
    today_open = DataFrame(index=price_index+Timedelta('9h30m'),
                           columns=tickers, data=price_data)
    today_open_1h = DataFrame(index=price_index+Timedelta('10h30m'),
                              columns=tickers, data=price_data)
    today_open_1h += today_open_1h*0.001
    today_open_3h = DataFrame(index=price_index+Timedelta('12h30m'),
                              columns=tickers, data=price_data)
    today_open_3h -= today_open_3h*0.002
    intraday_prices = concat([today_open, today_open_1h, today_open_3h]) \
        .sort_index()

    intraday_factor = DataFrame(index=factor_index+Timedelta('9h30m'),
                                columns=tickers, data=factor_data).stack()

    #
    # event factor
    #
    bevent_factor = DataFrame(index=bfactor_index, columns=tickers,
                              data=event_data).stack()

    event_factor = DataFrame(index=factor_index, columns=tickers,
                             data=event_data).stack()

    all_prices = [prices, bprices]
    all_factors = [factor, bfactor]
    all_events = [event_factor, bevent_factor]

    def __localize_prices_and_factor(self, prices, factor, tz):
        if tz is not None:
            factor = factor.unstack()
            factor.index = factor.index.tz_localize(tz)
            factor = factor.stack()
            prices = prices.copy()
            prices.index = prices.index.tz_localize(tz)
        return prices, factor

    @parameterized.expand([(2, (1, 5, 10), None),
                           (3, (2, 4, 6), 20)])
    def test_create_returns_tear_sheet(
            self,
            quantiles,
            periods,
            filter_zscore):
        """
        Test no exceptions are thrown
        """
        factor_data = get_clean_factor_and_forward_returns(
            self.factor,
            self.prices,
            quantiles=quantiles,
            periods=periods,
            filter_zscore=filter_zscore)

        create_returns_tear_sheet(
            factor_data, long_short=False, group_neutral=False, by_group=False)

    @parameterized.expand([(1, (1, 5, 10), None),
                           (4, (1, 2, 3, 7), 20)])
    def test_create_information_tear_sheet(
            self, quantiles, periods, filter_zscore):
        """
        Test no exceptions are thrown
        """
        factor_data = get_clean_factor_and_forward_returns(
            self.factor,
            self.prices,
            quantiles=quantiles,
            periods=periods,
            filter_zscore=filter_zscore)

        create_information_tear_sheet(
            factor_data, group_neutral=False, by_group=False)

    @parameterized.expand([
        (2, (2, 3, 6), None, 20),
        (4, (1, 2, 3, 7), None, None),
        (2, (2, 3, 6), ['1D', '2D'], 20),
        (4, (1, 2, 3, 7), ['1D'], None),
    ])
    def test_create_turnover_tear_sheet(
            self,
            quantiles,
            periods,
            turnover_periods,
            filter_zscore):
        """
        Test no exceptions are thrown
        """
        factor_data = get_clean_factor_and_forward_returns(
            self.factor,
            self.prices,
            quantiles=quantiles,
            periods=periods,
            filter_zscore=filter_zscore)

        create_turnover_tear_sheet(factor_data, turnover_periods)

    @parameterized.expand([(2, (1, 5, 10), None),
                           (3, (1, 2, 3, 7), 20)])
    def test_create_summary_tear_sheet(
            self,
            quantiles,
            periods,
            filter_zscore):
        """
        Test no exceptions are thrown
        """
        factor_data = get_clean_factor_and_forward_returns(
            self.factor,
            self.prices,
            quantiles=quantiles,
            periods=periods,
            filter_zscore=filter_zscore)

        create_summary_tear_sheet(
            factor_data, long_short=True, group_neutral=False)
        create_summary_tear_sheet(
            factor_data, long_short=False, group_neutral=False)

    @parameterized.expand([
        (2, (1, 5, 10), None, None),
        (3, (2, 4, 6), 20, 'US/Eastern'),
        (4, (1, 8), 20, None),
        (4, (1, 2, 3, 7), None, 'US/Eastern'),
    ])
    def test_create_full_tear_sheet(
            self,
            quantiles,
            periods,
            filter_zscore,
            tz):
        """
        Test no exceptions are thrown
        """
        for factor, prices in zip(self.all_factors, self.all_prices):

            prices, factor = self.__localize_prices_and_factor(prices,
                                                               factor,
                                                               tz)
            factor_data = get_clean_factor_and_forward_returns(
                factor,
                prices,
                groupby=self.factor_groups,
                quantiles=quantiles,
                periods=periods,
                filter_zscore=filter_zscore)

            create_full_tear_sheet(factor_data, long_short=False,
                                   group_neutral=False, by_group=False)
            create_full_tear_sheet(factor_data, long_short=True,
                                   group_neutral=False, by_group=True)
            create_full_tear_sheet(factor_data, long_short=True,
                                   group_neutral=True, by_group=True)

    @parameterized.expand([(2, (1, 5, 10), None, None),
                           (3, (2, 4, 6), 20, None),
                           (4, (3, 4), None, 'US/Eastern'),
                           (1, (2, 3, 6, 9), 20, 'US/Eastern')])
    def test_create_event_returns_tear_sheet(
            self, quantiles, periods, filter_zscore, tz):
        """
        Test no exceptions are thrown
        """
        for factor, prices in zip(self.all_factors, self.all_prices):

            prices, factor = self.__localize_prices_and_factor(prices,
                                                               factor,
                                                               tz)
            factor_data = get_clean_factor_and_forward_returns(
                factor,
                prices,
                groupby=self.factor_groups,
                quantiles=quantiles,
                periods=periods,
                filter_zscore=filter_zscore)

            create_event_returns_tear_sheet(factor_data, prices, avgretplot=(
                5, 11), long_short=False, group_neutral=False, by_group=False)
            create_event_returns_tear_sheet(factor_data, prices, avgretplot=(
                5, 11), long_short=True, group_neutral=False, by_group=False)
            create_event_returns_tear_sheet(factor_data, prices, avgretplot=(
                5, 11), long_short=False, group_neutral=True, by_group=False)
            create_event_returns_tear_sheet(factor_data, prices, avgretplot=(
                5, 11), long_short=False, group_neutral=False, by_group=True)
            create_event_returns_tear_sheet(factor_data, prices, avgretplot=(
                5, 11), long_short=True, group_neutral=False, by_group=True)
            create_event_returns_tear_sheet(factor_data, prices, avgretplot=(
                5, 11), long_short=False, group_neutral=True, by_group=True)

    @parameterized.expand([((6, 8), None, None),
                           ((6, 8), None, None),
                           ((6, 3), 20, None),
                           ((6, 3), 20, 'US/Eastern'),
                           ((0, 3), None, None),
                           ((3, 0), 20, 'US/Eastern')])
    def test_create_event_study_tear_sheet(
            self, avgretplot, filter_zscore, tz):
        """
        Test no exceptions are thrown
        """
        for factor, prices in zip(self.all_events, self.all_prices):

            prices, factor = self.__localize_prices_and_factor(prices,
                                                               factor,
                                                               tz)
            factor_data = get_clean_factor_and_forward_returns(
                factor, prices, bins=1, quantiles=None, periods=(
                    1, 2), filter_zscore=filter_zscore)

            create_event_study_tear_sheet(
                factor_data, prices, avgretplot=avgretplot)



from ..util import DLmodel
import torch
import torch.nn as nn
import torch.nn.functional as F
from math import log






class PositionalEncoding(nn.Module):
    """
    位置编码
    """
    def __init__(self, d_model, max_len=1000):
        super(PositionalEncoding, self).__init__()
        # 创建一个足够长的P
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return x
    

class BlockTransformer(nn.Module):
    """
    Transformer模块的编码器，一层全连接网络+编码器，实现(batch_size,seq_len,input_dim)->(batch_size,seq_len,d_model)的输出
    """
    def __init__(self, input_dim, d_model, num_heads, num_layers, dim_feedforward,drop_rate):
        super(BlockTransformer, self).__init__()
        
        self.d_model = d_model
        self.aft_dropout = nn.Dropout(drop_rate)
        self.pre_layer = nn.Linear(input_dim,d_model)
        self.pre_dropout = nn.Dropout(drop_rate)
        
        # 编码器层
        self.pos_encoder = PositionalEncoding(self.d_model, max_len=500)
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=self.d_model,
            nhead=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=drop_rate,
            activation='relu',
            batch_first=True,
            # attn_weights_need=attn_weights_need
        )
        
        # 编码器
        self.encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=num_layers)
        
    def forward(self, x):
        # x: (batch_size, seq_len, input_dim)
        
        ### 添加位置编码
        x = self.pre_layer(x)
        x = self.pre_dropout(x)
        x = self.pos_encoder(x.transpose(0,1)).transpose(0,1)
        
        x = self.encoder(x)
        
        # x:(batch_size, seq_len, d_model)
        return x

class BlockLSTM(nn.Module):
    """
    LSTM编码器模块，一层lstm编码器，实现(batch_siez,seq_len,input_dim)->(batch_size,seq_len,hidden_dim)的输出
    """
    def __init__(self, input_dim, hidden_dim, num_layers=1,lstm_drop=0):
        super(BlockLSTM, self).__init__()
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        # LSTM层
        self.lstm = nn.LSTM(input_dim, hidden_dim, num_layers, batch_first=True,dropout=lstm_drop)
        
    def forward(self, x):
        # 初始化隐藏状态和细胞状态
        h0 = torch.zeros((self.num_layers, x.size(0), self.hidden_dim)).to(x.device) 
        c0 = torch.zeros((self.num_layers, x.size(0), self.hidden_dim)).to(x.device)
        # 前向传播LSTM
        out, _ = self.lstm(x, (h0, c0))  # out: tensor of shape (batch_size, seq_length, hidden_dim)
        return out
    

class AttentionLSTM(nn.Module):
    def __init__(self,input_dim,d_model,num_heads,trans_layers,dim_feedforward,hidden_dim,lstm_layers,out_dim,backward,drop_rate):
        super(AttentionLSTM, self).__init__()
        self.backward = backward
        self.hidden_dim = hidden_dim
        self.trans_encoder = BlockTransformer(input_dim,d_model,num_heads,trans_layers,dim_feedforward,drop_rate)
        self.lstm_encoder = BlockLSTM(d_model,hidden_dim,lstm_layers)
        self.aft_layer1 = nn.Linear(backward*hidden_dim,300)
        self.drop_out1 = nn.Dropout(drop_rate)
        self.aft_layer2 = nn.Linear(300,100)
        self.drop_out2 = nn.Dropout(drop_rate)
        self.aft_layer3 = nn.Linear(100,out_dim)
        
    
    def forward(self,x):
        x = self.trans_encoder(x)
        x = self.lstm_encoder(x)
        x = x[:,-self.backward:,:].reshape(-1,self.backward*self.hidden_dim)
        x = F.relu(self.aft_layer1(x))
        x = self.drop_out1(x)
        x = F.relu(self.aft_layer2(x))
        x = self.drop_out2(x)
        x = self.aft_layer3(x)
        return x
        
    


import abc
from functools import reduce

import numpy as np
import pandas as pd
import sunlandsdatasdk as sd


# region Auxiliary functions
def returns(df):
    return df.rolling(2).apply(lambda x: x.iloc[-1] / x.iloc[0]) - 1


def ts_sum(df, window=10):
    """
    Wrapper function to estimate rolling sum.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series min over the past 'window' days.
    """
    return df.rolling(window).sum()


def sma(df, window=10):
    """
    Wrapper function to estimate SMA.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series min over the past 'window' days.
    """
    return df.rolling(window).mean()


def stddev(df, window=10):
    """
    Wrapper function to estimate rolling standard deviation.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series min over the past 'window' days.
    """
    return df.rolling(window).std()


def correlation(x, y, window=10):
    """
    Wrapper function to estimate rolling corelations.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series min over the past 'window' days.
    """
    return x.rolling(window).corr(y)


def covariance(x, y, window=10):
    """
    Wrapper function to estimate rolling covariance.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series min over the past 'window' days.
    """
    return x.rolling(window).cov(y)


def rolling_rank(na):
    """
    Auxiliary function to be used in pd.rolling_apply
    :param na: numpy array.
    :return: The rank of the close value in the array.
    """
    return rankdata(na, method='min')[-1]


def ts_rank(df, window=10):
    """
    Wrapper function to estimate rolling rank.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series rank over the past window days.
    """
    return df.rolling(window).apply(rolling_rank)


def rolling_prod(na):
    """
    Auxiliary function to be used in pd.rolling_apply
    :param na: numpy array.
    :return: The product of the values in the array.
    """
    return np.prod(na)


def product(df, window=10):
    """
    Wrapper function to estimate rolling product.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series product over the past 'window' days.
    """
    return df.rolling(window).apply(rolling_prod)


def ts_min(df, window=10):
    """
    Wrapper function to estimate rolling min.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series min over the past 'window' days.
    """
    return df.rolling(window).min()


def ts_max(df, window=10):
    """
    Wrapper function to estimate rolling min.
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: a pandas DataFrame with the time-series max over the past 'window' days.
    """
    return df.rolling(window).max()


def delta(df, period=1):
    """
    Wrapper function to estimate difference.
    :param df: a pandas DataFrame.
    :param period: the difference grade.
    :return: a pandas DataFrame with today’s value minus the value 'period' days ago.
    """
    return df.diff(period)


def delay(df, period=1):
    """
    Wrapper function to estimate lag.
    :param df: a pandas DataFrame.
    :param period: the lag grade.
    :return: a pandas DataFrame with lagged time series
    """
    return df.shift(period)


def rank(df):
    """
    Cross sectional rank
    :param df: a pandas DataFrame.
    :return: a pandas DataFrame with rank along columns.
    """
    return df.rank(axis=1, method='min', pct=True)
    # return df.rank(pct=True)


def scale(df, k=1):
    """
    Scaling time serie.
    :param df: a pandas DataFrame.
    :param k: scaling factor.
    :return: a pandas DataFrame rescaled df such that sum(abs(df)) = k
    """
    return df.mul(k).div(np.abs(df).sum())


def ts_argmax(df, window=10):
    """
    Wrapper function to estimate which day ts_max(df, window) occurred on
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: well.. that :)
    """
    return df.rolling(window).apply(np.argmax) + 1


def ts_argmin(df, window=10):
    """
    Wrapper function to estimate which day ts_min(df, window) occurred on
    :param df: a pandas DataFrame.
    :param window: the rolling window.
    :return: well.. that :)
    """
    return df.rolling(window).apply(np.argmin) + 1


def decay_linear(df, period=10):
    """
    Linear weighted moving average implementation.
    :param df: a pandas DataFrame.
    :param period: the LWMA period
    :return: a pandas DataFrame with the LWMA.
    """
    # Clean data
    if df.isnull().values.any():
        df.fillna(method='ffill', inplace=True)
        df.fillna(method='bfill', inplace=True)
        df.fillna(value=0, inplace=True)
    na_lwma = np.zeros_like(df)
    na_lwma[:period, :] = df.iloc[:period, :]
    na_series = df.as_matrix()

    divisor = period * (period + 1) / 2
    y = (np.arange(period) + 1) * 1.0 / divisor
    # Estimate the actual lwma with the actual close.
    # The backtest engine should assure to be snooping bias free.
    for row in range(period - 1, df.shape[0]):
        x = na_series[row - period + 1: row + 1, :]
        na_lwma[row, :] = (np.dot(x.T, y))
    return pd.DataFrame(na_lwma, index=df.index, columns=['CLOSE'])


class Factor(abc.ABC):
    """
    因子定义抽象基类，自定义因子类必须继承它，实现calc方式
    """
    # 因子的名称， 不能与基础因子冲突。
    name = ''
    # 获取数据的最长时间窗口，返回的是日级别的数据。
    max_window = 1
    # 依赖的基础因子名称，包含 'open', 'close', 'low', 'high', 'volume', 'vwap', 'returns' 字段
    dependencies = []

    @abc.abstractmethod
    def calc(self, data):
        """
        计算因子值，子类必须实现

        :param data 是pandas.DataFrame对象, 行索引是datetime.datetime对象; 列是dependencies中数据
        :return 返回pandas.DataFrame或pandas.Series, 行索引是datetime.datetime对象, 列是股票标的, 值数据是因子值
        """
        pass


def calc_factors(securities, factors, start_date, end_date, frequency='daily', fq='pre'):
    """
    计算一支或者多只股票的多个因子

    :param securities 一支证券代码或者一个证券代码的list
    :param factors 因子实例，Factor()子类实例可遍历序列
    :param start_date 字符串或者 datetime.datetime/datetime.date 对象, 开始时间
    :param end_date 格式同上, 结束时间, 默认是'2015-12-31', 包含此日期.
    :param frequency 单位时间长度, 几天或者几分钟, 现在支持'Xd','Xm', 'daily'(等同于'1d'), 'minute'(等同于'1m'), X是一个正整数, 分别表示X天和X分钟
    :param fq 'pre',
    :return 返回pandas.DataFrame对象, 行索引是pandas.MultiIndex二级, 一级是datetime.datetime对象, 二级是股票标的; 列索引是因子名称列表; 值数据是对应因子值
    """
    # 获取聚宽JoinQuant标的行情数据，需要切换的接口
    # todo 需要切换到尚德自定义的接口，获取数据
    price = sd.get_price(securities, start_date=start_date, end_date=end_date,
                      fields=["date","issue", "market", "time", "quoteType", "preclose" ,"open","high","low","close","numTrades","volume","value","adj"])

    # 因子日收益率 =(T+1收盘价/T收盘价) - 1
    price['returns'] = price['close'] / price['preclose'] - 1
    price['avg'] = price['value'] / price['volume']

    # 计算因子所需数据与结构：开盘价，收盘价，最低价，最高价，成交量，成交均价，日收益率
    data = price.pivot(index='date', columns='issue',
                       values=['open', 'close', 'low', 'high', 'volume', 'avg', 'returns'])


    data.rename(columns={"avg": "vwap"}, inplace=True)

    factor_results = []

    for factor in factors:
        if issubclass(factor.__class__, Factor):
            sub_data = data[factor.dependencies]
            factor_data = factor.calc(sub_data)
            factor_data = pd.DataFrame(factor_data).stack()
            factor_data.name = factor.name
            factor_results.append(factor_data)
        else:
            print(f"{factor.__class__} is not subclass of factor.Factor")

    multi_factor_result = reduce(lambda left, right: pd.merge(left, right, left_index=True, right_index=True),
                                 factor_results)
    # multi_factor_result = multi_factor_result.unstack()
    return multi_factor_result
